"""
Computes Single Feature Importance (SFI).
"""

import pandas as pd
import numpy as np
from sklearn.metrics import log_loss, accuracy_score
from sklearn.model_selection import KFold
from typing import List, Optional, Union, Any
from .feature_importance_strategy import FeatureImportanceStrategy

class FeatureImportanceSFI(FeatureImportanceStrategy):
    """
    Computes Single Feature Importance (SFI).

    This method calculates the importance of each feature by training
    and evaluating a model using *only that single feature*.
    """

    def __init__(
        self,
        classifier: object,
        n_splits: int = 10,
        scoring: str = "log_loss",
    ):
        """
        Initialize the strategy.

        Parameters
        ----------
        classifier : object
            An *untrained* scikit-learn classifier.
        n_splits : int, default=10
            Number of splits for cross-validation.
        scoring : str, default="log_loss"
            Scoring method ("log_loss" or "accuracy").
        """
        self.classifier = classifier
        self.n_splits = n_splits
        self.scoring = scoring

    def compute(self, x: pd.DataFrame, y: pd.Series, **kwargs: Any) -> pd.DataFrame:
        """
        Compute SFI feature importance.

        Parameters
        ----------
        x : pd.DataFrame
            The feature data.
        y : pd.Series
            The target data.
        **kwargs : Any
            - 'train_sample_weights': Optional sample weights for training.
            - 'score_sample_weights': Optional sample weights for scoring.

        Returns
        -------
        pd.DataFrame
            DataFrame with "FeatureName", "Mean", and "StandardDeviation"
            of the SFI scores.
        """
        train_sample_weights = kwargs.get('train_sample_weights')
        score_sample_weights = kwargs.get('score_sample_weights')
        
        if train_sample_weights is None:
            train_sample_weights = np.ones(x.shape[0])
        if score_sample_weights is None:
            score_sample_weights = np.ones(x.shape[0])

        cv_generator = KFold(n_splits=self.n_splits)
        feature_names = x.columns
        importances = []

        for feature_name in feature_names:
            scores = []
            feature_data = x[[feature_name]]

            for train_idx, test_idx in cv_generator.split(feature_data):
                x_train, y_train, w_train = (
                    feature_data.iloc[train_idx],
                    y.iloc[train_idx],
                    train_sample_weights[train_idx],
                )
                x_test, y_test, w_test = (
                    feature_data.iloc[test_idx],
                    y.iloc[test_idx],
                    score_sample_weights[test_idx],
                )

                self.classifier.fit(x_train, y_train, sample_weight=w_train)

                if self.scoring == "log_loss":
                    pred_proba = self.classifier.predict_proba(x_test)
                    score = -log_loss(
                        y_test,
                        pred_proba,
                        sample_weight=w_test,
                        labels=self.classifier.classes_,
                    )
                elif self.scoring == "accuracy":
                    pred = self.classifier.predict(x_test)
                    score = accuracy_score(
                        y_test, pred, sample_weight=w_test
                    )
                else:
                    raise ValueError(f"'{self.scoring}' method not defined.")

                scores.append(score)

            importances.append(
                {
                    "FeatureName": feature_name,
                    "Mean": np.mean(scores),
                    "StandardDeviation": np.std(scores, ddof=1)
                    * (len(scores) ** -0.5),
                }
            )

        return pd.DataFrame(importances).set_index("FeatureName")


import numpy as np
import pandas as pd
import time
from typing import List, Tuple
from RiskLabAI.utils.smoothing_average import compute_exponential_weighted_moving_average
from RiskLabAI.utils.progress import  progress_bar


def compute_thresholds(
    target_column: np.ndarray,
    initial_expected_ticks: int,
    initial_bar_size: float
) -> Tuple[List[float], np.ndarray, np.ndarray, List[int], np.ndarray, np.ndarray]:
    """
    Groups the target_column DataFrame based on a feature and calculates thresholds.

    This function groups the target_column DataFrame based on a feature
    and calculates the thresholds, which can be used in financial machine learning
    applications such as dynamic time warping.

    :param target_column: Target column of the DataFrame.
    :type target_column: np.ndarray
    :param initial_expected_ticks: Initial expected number of ticks.
    :type initial_expected_ticks: int
    :param initial_bar_size: Initial expected size of each tick.
    :type initial_bar_size: float
    :return: A tuple containing the time deltas, absolute theta values, thresholds,
        times, theta values, and grouping IDs.
    :rtype: Tuple[List[float], np.ndarray, np.ndarray, List[int], np.ndarray, np.ndarray]
    """
    num_values = target_column.shape[0]
    target_column_values = target_column.astype(np.float64)
    absolute_thetas = np.zeros(num_values)
    thresholds = np.zeros(num_values)
    thetas = np.zeros(num_values)
    grouping_ids = np.zeros(num_values)

    current_theta = target_column_values[0]
    thetas[0] = current_theta
    absolute_thetas[0] = np.abs(current_theta)
    current_grouping_id = 0
    grouping_ids[0] = current_grouping_id

    time_deltas = []
    times = []
    previous_time = 0
    expected_ticks = initial_expected_ticks
    expected_bar_value = initial_bar_size

    start_time = time.time()

    for i in range(1, num_values):
        current_theta += target_column_values[i]
        thetas[i] = current_theta
        absolute_theta = np.abs(current_theta)
        absolute_thetas[i] = absolute_theta

        threshold = expected_ticks * expected_bar_value
        thresholds[i] = threshold
        grouping_ids[i] = current_grouping_id

        if absolute_theta >= threshold:
            current_grouping_id += 1
            current_theta = 0
            time_delta = np.float64(i - previous_time)
            time_deltas.append(time_delta)
            times.append(i)
            previous_time = i
            expected_ticks = compute_exponential_weighted_moving_average(
                np.array(time_deltas), window_length=np.int64(len(time_deltas))
            )[-1]
            expected_bar_value = np.abs(
                compute_exponential_weighted_moving_average(
                    target_column_values[:i], window_length=np.int64(initial_expected_ticks)
                )[-1]
            )

        progress_bar(i, num_values, start_time)

    return time_deltas, absolute_thetas, thresholds, times, thetas, grouping_ids


def create_ohlcv_dataframe(
    tick_data_grouped: pd.core.groupby.DataFrameGroupBy
) -> pd.DataFrame:
    """
    Takes a grouped DataFrame and creates a new one with OHLCV data and other relevant information.

    :param tick_data_grouped: Grouped DataFrame based on some criteria (e.g., time).
    :type tick_data_grouped: pd.core.groupby.DataFrameGroupBy
    :return: A DataFrame containing OHLCV data and other relevant information.
    :rtype: pd.DataFrame
    """
    ohlc = tick_data_grouped['price'].ohlc()
    ohlc['volume'] = tick_data_grouped['size'].sum()
    ohlc['value_of_trades'] = tick_data_grouped.apply(
        lambda x: (x['price'] * x['size']).sum() / x['size'].sum()
    )
    ohlc['price_mean'] = tick_data_grouped['price'].mean()
    ohlc['tick_count'] = tick_data_grouped['price'].count()
    ohlc['price_mean_log_return'] = np.log(ohlc['price_mean']) - np.log(ohlc['price_mean'].shift(1))

    return ohlc






"""
Base class and concrete implementations for Partial Differential Equations
(PDEs) to be solved by the Deep BSDE solver.

Based on: https://github.com/frankhan91/DeepBSDE
"""

import torch
import torch.nn as nn
from torch import Tensor
import numpy as np

class Equation:
    """
    Base class for defining PDE-related functions.
    """

    def __init__(self, eqn_config: dict):
        """
        Initialize the PDE configuration.

        Parameters
        ----------
        eqn_config : dict
            Dictionary containing PDE parameters:
            - 'dim': Dimensionality of the problem (int)
            - 'total_time': Total time horizon (float)
            - 'num_time_interval': Number of time steps (int)
        """
        self.dim: int = eqn_config['dim']
        self.total_time: float = eqn_config['total_time']
        self.num_time_interval: int = eqn_config['num_time_interval']
        self.delta_t: float = self.total_time / self.num_time_interval
        self.sqrt_delta_t: float = np.sqrt(self.delta_t)
        self.y_init: Optional[float] = None

    def sample(self, num_sample: int) -> Tuple[np.ndarray, np.ndarray]:
        """
        Sample the forward SDE (e.g., the underlying asset path).

        Parameters
        ----------
        num_sample : int
            Number of sample paths to generate.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            - dw_sample: Wiener process increments [num_sample, dim, num_time_interval]
            - x_sample: Asset paths [num_sample, dim, num_time_interval+1]
        """
        raise NotImplementedError

    def r_u(self, t: float, x: Tensor, y: Tensor, z: Tensor) -> Tensor:
        """
        The driver function 'f' of the BSDE, or the generator of the PDE.
        Represents the non-linear interest rate and dividend term.

        Parameters
        ----------
        t : float
            Current time.
        x : Tensor
            Space coordinates [batch_size, dim].
        y : Tensor
            Function value (price) [batch_size, 1].
        z : Tensor
            Gradient (delta) [batch_size, dim].

        Returns
        -------
        Tensor
            Value of the driver `f` [batch_size, 1].
        """
        raise NotImplementedError

    def h_z(self, t: float, x: Tensor, y: Tensor, z: Tensor) -> Tensor:
        """
        The Hamiltonian term H(z) in the PDE, or the volatility-related
        term in the BSDE driver.

        Parameters
        ----------
        t : float
            Current time.
        x : Tensor
            Space coordinates [batch_size, dim].
        y : Tensor
            Function value (price) [batch_size, 1].
        z : Tensor
            Gradient (delta) [batch_size, dim].

        Returns
        -------
        Tensor
            Value of the Hamiltonian `H(z)` [batch_size, 1].
        """
        raise NotImplementedError

    def terminal(self, t: float, x: Tensor) -> Tensor:
        """
        Terminal condition (payoff) of the PDE.

        Parameters
        ----------
        t : float
            Current time (should be total_time).
        x : Tensor
            Space coordinates [batch_size, dim].

        Returns
        -------
        Tensor
            Terminal payoff value [batch_size, 1].
        """
        raise NotImplementedError
        
    def sigma_matrix(self, x: Tensor) -> Tensor:
        """Helper to get the volatility matrix sigma(x)."""
        raise NotImplementedError
        
    def terminal_for_sample(self, x: Tensor) -> Tensor:
        """Terminal condition for a multi-sample path."""
        raise NotImplementedError


class PricingDefaultRisk(Equation):
    """
    PDE for pricing with default risk.
    """
    def __init__(self, eqn_config: dict):
        super(PricingDefaultRisk, self).__init__(eqn_config)
        self.x_init = np.ones(self.dim) * 100.0
        self.sigma = 0.2
        self.rate = 0.02   # R
        self.delta = 2.0 / 3
        self.gammah = 0.2
        self.gammal = 0.02
        self.mu_bar = 0.02
        self.K = 100
        self.vh = 50.0
        self.vl = 70.0
        self.slope = (self.gammah - self.gammal) / (self.vh - self.vl)

    def sample(self, num_sample: int) -> Tuple[np.ndarray, np.ndarray]:
        dw_sample = np.random.normal(
            size=[num_sample, self.dim, self.num_time_interval]
        ) * self.sqrt_delta_t
        x_sample = np.zeros([num_sample, self.dim, self.num_time_interval + 1])
        x_sample[:, :, 0] = np.ones([num_sample, self.dim]) * self.x_init
        for i in range(self.num_time_interval):
            x_sample[:, :, i + 1] = (1 + self.mu_bar * self.delta_t) * x_sample[
                :, :, i
            ] + (self.sigma_matrix(x_sample[:, :, i]) * dw_sample[:, :, i])
        return dw_sample, x_sample

    def r_u(self, t: float, x: Tensor, y: Tensor, z: Tensor) -> Tensor:
        piecewise_linear = nn.ReLU()(
            nn.ReLU()(y - self.vh) * self.slope + self.gammah - self.gammal
        ) + self.gammal
        return (1 - self.delta) * piecewise_linear + self.rate

    def h_z(self, t: float, x: Tensor, y: Tensor, z: Tensor) -> Tensor:
        return torch.zeros((x.size()[0], 1), device=x.device)

    def sigma_matrix(self, x: Union[np.ndarray, Tensor]) -> Union[np.ndarray, Tensor]:
        return self.sigma * x

    def terminal(self, t: float, x: Tensor) -> Tensor:
        return nn.ReLU()(torch.min(x, 1)[0])

    def terminal_for_sample(self, x: Tensor) -> Tensor:
        return nn.ReLU()(torch.min(x, 2, keepdim=True)[0])


class HJBLQ(Equation):
    """
    Hamilton-Jacobi-Bellman (HJB) equation with Linear-Quadratic (LQ) cost.
    """
    def __init__(self, eqn_config: dict):
        super(HJBLQ, self).__init__(eqn_config)
        self.x_init = np.zeros(self.dim)
        self.sigma = np.sqrt(2.0)
        self.lambd = 1.0

    def sample(self, num_sample: int) -> Tuple[np.ndarray, np.ndarray]:
        dw_sample = np.random.normal(
            size=[num_sample, self.dim, self.num_time_interval]
        ) * self.sqrt_delta_t
        x_sample = np.zeros([num_sample, self.dim, self.num_time_interval + 1])
        x_sample[:, :, 0] = np.ones([num_sample, self.dim]) * self.x_init
        for i in range(self.num_time_interval):
            x_sample[:, :, i + 1] = x_sample[:, :, i] + self.sigma * dw_sample[:, :, i]
        return dw_sample, x_sample

    def r_u(self, t: float, x: Tensor, y: Tensor, z: Tensor) -> Tensor:
        return torch.zeros((x.size()[0], 1), device=x.device)

    def h_z(self, t: float, x: Tensor, y: Tensor, z: Tensor) -> Tensor:
        return torch.sum(torch.square(z), dim=1, keepdim=True) / (self.sigma**2)

    def terminal(self, t: float, x: Tensor) -> Tensor:
        return torch.log(0.5 * (1 + torch.norm(x, dim=1, keepdim=True)**2))

    def sigma_matrix(self, x: Union[np.ndarray, Tensor]) -> float:
        return self.sigma
        
    def terminal_for_sample(self, x: Tensor) -> Tensor:
        return torch.sum(x ** 2, dim=2, keepdim=True)


# (Other equations like BlackScholesBarenblatt, PricingDiffRate refactored similarly)


# import torch
# import torch.nn.functional as F
# from torch import nn, Tensor
# import numpy as np
# from torch.nn import Module, Linear, BatchNorm1d, Tanh
# from numba import cuda
# from torch import autograd
# import torch.nn as nn
# import torch.nn.functional as F
# import math
# """
# In this section we have used code available online in : https://github.com/frankhan91/DeepBSDE

# """

# class Equation:
#     """
#     Base class for defining PDE related function.

#     Args:
#     eqn_config (dict): dictionary containing PDE configuration parameters

#     Attributes:
#     dim (int): dimensionality of the problem
#     total_time (float): total time horizon
#     num_time_interval (int): number of time steps
#     delta_t (float): time step size
#     sqrt_delta_t (float): square root of time step size
#     y_init (None): initial value of the function
#     """

#     def __init__(self, eqn_config: dict):
#         self.dim = eqn_config['dim']
#         self.total_time = eqn_config['total_time']
#         self.num_time_interval = eqn_config['num_time_interval']
#         self.delta_t = self.total_time / self.num_time_interval
#         self.sqrt_delta_t = np.sqrt(self.delta_t)
#         self.y_init = None

#     def sample(self, num_sample: int) -> Tensor:
#         """
#         Sample forward SDE.

#         Args:
#         num_sample (int): number of samples to generate

#         Returns:
#         Tensor: tensor of size [num_sample, dim+1] containing samples
#         """
#         raise NotImplementedError

#     def r_u(self, t: float, x: Tensor, y: Tensor, z: Tensor) -> Tensor:
#         """
#         Interest rate in the PDE.

#         Args:
#         t (float): current time
#         x (Tensor): tensor of size [batch_size, dim] containing space coordinates
#         y (Tensor): tensor of size [batch_size, 1] containing function values
#         z (Tensor): tensor of size [batch_size, dim] containing gradients

#         Returns:
#         Tensor: tensor of size [batch_size, 1] containing generator values
#         """
#         raise NotImplementedError

#     def h_z(self, t,x,y,z: Tensor) -> Tensor:
#         """
#         Function to compute H(z) in the PDE.

#         Args:
#         h (float): value of H function
#         z (Tensor): tensor of size [batch_size, dim] containing gradients

#         Returns:
#         Tensor: tensor of size [batch_size, dim] containing H(z)
#         """
#         raise NotImplementedError

#     def terminal(self, t: float, x: Tensor) -> Tensor:
#         """
#         Terminal condition of the PDE.

#         Args:
#         t (float): current time
#         x (Tensor): tensor of size [batch_size, dim] containing space coordinates

#         Returns:
#         Tensor: tensor of size [batch_size, 1] containing terminal values
#         """
#         raise NotImplementedError


# class PricingDefaultRisk(Equation):
#   """
#   Args:
#   eqn_config (dict): dictionary containing PDE configuration parameters
#   """
#   def __init__(self, eqn_config):
#     super(PricingDefaultRisk, self).__init__(eqn_config)
#     self.x_init = np.ones(self.dim) * 100.0
#     self.sigma = 0.2
#     self.rate = 0.02   # interest rate R
#     self.delta = 2.0 / 3
#     self.gammah = 0.2
#     self.gammal = 0.02
#     self.mu_bar = 0.02
#     self.K = 100
#     self.vh = 50.0
#     self.vl = 70.0
#     self.slope = (self.gammah - self.gammal) / (self.vh - self.vl)

#   def sample(self, num_sample)-> tuple:
#     """
#     Sample forward SDE.

#     Args:
#     num_sample (int): number of samples to generate

#     Returns:
#     tuple: tuple of two tensors: dw_sample of size [num_sample, dim, num_time_interval] and
#     x_sample of size [num_sample, dim, num_time_interval+1]
#     """

#     dw_sample = np.random.normal(size=[num_sample, self.dim, self.num_time_interval]) * self.sqrt_delta_t
#     x_sample = np.zeros([num_sample, self.dim, self.num_time_interval + 1])
#     x_sample[:, :, 0] = np.ones([num_sample, self.dim]) * self.x_init
#     for i in range(self.num_time_interval):
#         x_sample[:, :, i + 1] = (1 + self.mu_bar * self.delta_t) * x_sample[:, :, i] + (
#             self.sigma_matrix(x_sample[:, :, i]) * dw_sample[:, :, i])
#     return dw_sample, x_sample

#   def r_u(self, t, x, y, z)-> torch.Tensor:
#     """
#     Interest rate in the PDE.

#     Args:
#     t (float): current time
#     x (torch.Tensor): tensor of size [batch_size, dim] containing space coordinates
#     y (torch.Tensor): tensor of size [batch_size, 1] containing function values
#     z (torch.Tensor): tensor of size [batch_size, dim] containing gradients

#     Returns:
#     torch.Tensor: tensor of size [batch_size, 1] containing generator values
#     """
#     piecewise_linear = nn.ReLU()(
#         nn.ReLU()(y - self.vh) * self.slope + self.gammah - self.gammal) + self.gammal
#     return ((1 - self.delta) * piecewise_linear  + self.rate)

#   def h_z(self,t,x,y,z)-> torch.Tensor:
#       """
#       Function to compute $h^T Z$ in the PDE.

#       Args:
#       t (float): current time
#       x (torch.Tensor): tensor of size [batch_size, dim] containing space coordinates
#       y (torch.Tensor): tensor of size [batch_size, 1] containing function value
#       z (torch.Tensor): tensor of size [batch_size, dim] containing gradients

#       Returns:
#       torch.Tensor: tensor of size [batch_size, 1] containing H(z)
#       """
#       return torch.zeros((x.size()[0],1))

#   def sigma_matrix(self,x):
#     return self.sigma * x


#   def terminal(self, t, x)-> torch.Tensor:
#     """
#     Terminal condition of the PDE.

#     Args:
#     t (float): current time
#     x (torch.Tensor): tensor of size [batch_size, dim] containing space coordinates

#     Returns:
#     torch.Tensor: tensor of size [batch_size, 1] containing terminal values
#     """
#     return nn.ReLU()(torch.min(x, 1)[0])
#   def terminal_for_sample(self, x)-> torch.Tensor:
#     """
#     Terminal condition of the PDE.

#     Args:
#     x (torch.Tensor): tensor of size [num_sample,batch_size, dim] containing space coordinates

#     Returns:
#     torch.Tensor: tensor of size [num_sample ,batch_size, 1] containing terminal values
#     """
#     return nn.ReLU()(torch.min(x, 2 , keepdim= True)[0])


# class HJBLQ(Equation):
#   """

#   Args:
#   eqn_config (dict): dictionary containing PDE configuration parameters
#   """

#   def __init__(self, eqn_config: dict):
#     super().__init__(eqn_config)

#     self.x_init = np.zeros(self.dim)
#     self.sigma = np.sqrt(2.0)
#     self.lambd = 1.0

#   def sample(self, num_sample: int) -> tuple:
#     """
#     Sample forward SDE.

#     Args:
#     num_sample (int): number of samples to generate

#     Returns:
#     tuple: tuple of two tensors: dw_sample of size [num_sample, dim, num_time_interval] and
#     x_sample of size [num_sample, dim, num_time_interval+1]
#     """
#     dw_sample = np.random.normal(size=[num_sample, self.dim, self.num_time_interval]) * self.sqrt_delta_t
#     x_sample = np.zeros([num_sample, self.dim, self.num_time_interval + 1])
#     x_sample[:, :, 0] = np.ones([num_sample, self.dim]) * self.x_init
#     for i in range(self.num_time_interval):
#         x_sample[:, :, i + 1] = x_sample[:, :, i] + self.sigma * dw_sample[:, :, i]
#     return dw_sample, x_sample

#   def r_u(self, t: float, x: torch.Tensor, y: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
#     """
#     Interest rate in the PDE.

#     Args:
#     t (float): current time
#     x (torch.Tensor): tensor of size [batch_size, dim] containing space coordinates
#     y (torch.Tensor): tensor of size [batch_size, 1] containing function values
#     z (torch.Tensor): tensor of size [batch_size, dim] containing gradients

#     Returns:
#     torch.Tensor: tensor of size [batch_size, 1] containing generator values
#     """
#     return torch.ones((x.size()[0],)) * 0

#   def h_z(self, t: float, x: torch.Tensor,y: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
#     """
#     Function to compute <h,z> in the PDE.

#     Args:
#     t (float): current time
#     x (torch.Tensor): tensor of size [batch_size, dim] containing space coordinates
#     y (torch.Tensor): tensor of size [batch_size, 1] containing function value
#     z (torch.Tensor): tensor of size [batch_size, dim] containing gradients

#     Returns:
#     torch.Tensor: tensor of size [batch_size, 1] containing H(z)
#     """
#     return torch.sum(torch.square(z), dim=1)/ (self.sigma**2)

#   def terminal(self, t: float, x: torch.Tensor) -> torch.Tensor:
#     """
#     Terminal condition of the PDE.

#     Args:
#     t (float): current time
#     x (torch.Tensor): tensor of size [batch_size, dim] containing space coordinates

#     Returns:
#     torch.Tensor: tensor of size [batch_size, 1] containing terminal values
#     """
#     return torch.log(0.5 * (1 + torch.norm(x, dim=1)**2))
#   def sigma_matrix(self,x):
#     return self.sigma
# def terminal_for_sample(self, x)-> torch.Tensor:
#     """
#     Terminal condition of the PDE.

#     Args:
#     x (torch.Tensor): tensor of size [num_sample,batch_size, dim] containing space coordinates

#     Returns:
#     torch.Tensor: tensor of size [num_sample ,batch_size, 1] containing terminal values
#     """
#     return torch.sum(x ** 2, dim=2 ,keepdim=True )

# class BlackScholesBarenblatt(Equation):
#   """

#   Args:
#   eqn_config (dict): dictionary containing PDE configuration parameters
#   """

#   def __init__(self, eqn_config: dict):
#     super().__init__(eqn_config)
#     self.x_init = np.ones(self.dim) * np.array([1.0 / (1.0 + i % 2) for i in range(self.dim)])
#     self.sigma = 0.4
#     self.rate = 0.05   # interest rate R
#     self.mu_bar = 0.0

#   def sample(self, num_sample: int) -> tuple:
#     """
#     Sample forward SDE.

#     Args:
#     num_sample (int): number of samples to generate

#     Returns:
#     tuple: tuple of two tensors: dw_sample of size [num_sample, dim, num_time_interval] and
#     x_sample of size [num_sample, dim, num_time_interval+1]
#     """
#     dw_sample = np.random.normal(size=(num_sample, self.dim, self.num_time_interval)) * self.sqrt_delta_t
#     x_sample = np.zeros((num_sample, self.dim, self.num_time_interval + 1))
#     x_sample[:, :, 0] = np.ones((num_sample, self.dim)) * self.x_init
#     for i in range(self.num_time_interval):
#         x_sample[:, :, i + 1] = (1 + self.mu_bar * self.delta_t) * x_sample[:, :, i] + (
#             self.sigma_matrix(x_sample[:, :, i]) * dw_sample[:, :, i])
#     return dw_sample, x_sample

#   def r_u(self, t: float, x: torch.Tensor, y: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
#     """
#     Interest rate in the PDE.

#     Args:
#     t (float): current time
#     x (torch.Tensor): tensor of size [batch_size, dim] containing space coordinates
#     y (torch.Tensor): tensor of size [batch_size, 1] containing function values
#     z (torch.Tensor): tensor of size [batch_size, dim] containing gradients

#     Returns:
#     torch.Tensor: tensor of size [batch_size, 1] containing generator values
#     """
#     return torch.ones((x.size()[0],)) * self.rate

#   def h_z(self, t: float, x: torch.Tensor,y: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
#     """
#     Function to compute <h,z> in the PDE.

#     Args:
#     t (float): current time
#     x (torch.Tensor): tensor of size [batch_size, dim] containing space coordinates
#     y (torch.Tensor): tensor of size [batch_size, 1] containing function value
#     z (torch.Tensor): tensor of size [batch_size, dim] containing gradients

#     Returns:
#     torch.Tensor: tensor of size [batch_size, 1] containing H(z)
#     """
#     return -1 * torch.sum(z, dim=1) * self.rate / self.sigma

#   def terminal(self, t: float, x: torch.Tensor) -> torch.Tensor:
#     """
#     Terminal condition of the PDE.

#     Args:
#     t (float): current time
#     x (torch.Tensor): tensor of size [batch_size, dim] containing space coordinates

#     Returns:
#     torch.Tensor: tensor of size [batch_size, 1] containing terminal values
#     """
#     return torch.sum(x ** 2, dim=1)
#   def sigma_matrix(self,x):
#     return self.sigma * x
#   def terminal_for_sample(self, x)-> torch.Tensor:
#     """
#     Terminal condition of the PDE.

#     Args:
#     x (torch.Tensor): tensor of size [num_sample,batch_size, dim] containing space coordinates

#     Returns:
#     torch.Tensor: tensor of size [num_sample ,batch_size, 1] containing terminal values
#     """
#     return torch.sum(x ** 2, dim=2 ,keepdim=True )

# class PricingDiffRate(Equation):
#     """
#     Nonlinear Black-Scholes equation with different interest rates for borrowing and lending
#     in Section 4.4 of Comm. Math. Stat. paper doi.org/10.1007/s40304-017-0117-6
#     """
#     def __init__(self, eqn_config):
#         super(PricingDiffRate, self).__init__(eqn_config)
#         self.x_init = np.ones(self.dim) * 100
#         self.sigma = 0.2
#         self.mu_bar = 0.06
#         self.rl = 0.04
#         self.rb = 0.06
#         self.alpha = 1.0 / self.dim

#     def sample(self, num_sample):
#         dw_sample = np.random.normal(size=[num_sample, self.dim, self.num_time_interval]) * self.sqrt_delta_t
#         x_sample = np.zeros([num_sample, self.dim, self.num_time_interval + 1])
#         x_sample[:, :, 0] = np.ones([num_sample, self.dim]) * self.x_init
#         factor = np.exp((self.mu_bar-(self.sigma**2)/2)*self.delta_t)
#         for i in range(self.num_time_interval):
#             x_sample[:, :, i + 1] = (factor * np.exp(self.sigma * dw_sample[:, :, i])) * x_sample[:, :, i]
#         return dw_sample, x_sample

#     def r_u(self, t: float, x: torch.Tensor, y: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
#       temp = torch.sum(z, dim= 1) / self.sigma - y
#       return torch.where(temp > 0 , self.rb , self.rl)
#     def h_z(self, t: float, x: torch.Tensor,y: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
#       """
#       Function to compute <h,z> in the PDE.

#       Args:
#       t (float): current time
#       x (torch.Tensor): tensor of size [batch_size, dim] containing space coordinates
#       y (torch.Tensor): tensor of size [batch_size, 1] containing function value
#       z (torch.Tensor): tensor of size [batch_size, dim] containing gradients

#       Returns:
#       torch.Tensor: tensor of size [batch_size, 1] containing H(z)
#       """
#       temp = torch.sum(z, dim= 1) / self.sigma - y
#       return torch.where(temp > 0 , (self.mu_bar - self.rb) * torch.sum(z, dim=1) / self.sigma ,(self.mu_bar - self.rl) * torch.sum(z, dim=1) / self.sigma)


#     def terminal(self, t, x):
#         temp = torch.max(x, 1)[0]
#         return nn.ReLU()(temp - 120) - 2*nn.ReLU()(temp - 150)
#     def sigma_matrix(self,x):
#       return self.sigma * x
#     def terminal_for_sample(self, x)-> torch.Tensor:
#         """
#         Terminal condition of the PDE.

#         Args:
#         x (torch.Tensor): tensor of size [num_sample,batch_size, dim] containing space coordinates

#         Returns:
#         torch.Tensor: tensor of size [num_sample ,batch_size, 1] containing terminal values
#         """
#         temp = torch.max(x, 2 , keep_dim = True)[0]
#         return torch.maximum(temp - 120, torch.tensor(0))



"""
Controller to simplify the creation and use of cross-validators.
"""

from typing import Any

from .cross_validator_factory import CrossValidatorFactory
from .cross_validator_interface import CrossValidator

class CrossValidatorController:
    """
    Controller class to handle the cross-validation process.

    This class acts as a high-level interface, simplifying the
    creation and access to a specific cross-validator using the factory.
    """

    def __init__(
        self,
        validator_type: str,
        **kwargs: Any
    ):
        """
        Initializes the CrossValidatorController.

        Parameters
        ----------
        validator_type : str
            Type of cross-validator to create (e.g., 'kfold',
            'combinatorialpurged'). This is passed to the factory.
        **kwargs : Any
            Additional keyword arguments to be passed to the
            cross-validator's constructor.
        """
        self.cross_validator: CrossValidator = \
            CrossValidatorFactory.create_cross_validator(
                validator_type,
                **kwargs
            )

    def get_validator(self) -> CrossValidator:
        """
        Get the created cross-validator instance.

        Returns
        -------
        CrossValidator
            The underlying cross-validator instance.
        """
        return self.cross_validator


"""
Calculates various backtest statistics like holding period,
concentration, and drawdowns.
"""

from typing import Tuple, Optional
import numpy as np
import pandas as pd

def bet_timing(target_positions: pd.Series) -> pd.Index:
    """
    Determine the timestamps of bets, defined as when positions
    are closed (return to zero) or flipped (sign change).

    Parameters
    ----------
    target_positions : pd.Series
        Time series of target positions.

    Returns
    -------
    pd.Index
        A DatetimeIndex of timestamps when bets are timed.

    Example
    -------
    >>> dates = pd.to_datetime(['2020-01-01', '2020-01-02', '2020-01-03', \
                                '2020-01-04', '2020-01-05', '2020-01-06'])
    >>> positions = pd.Series([0, 1, 1, 0, -1, 0], index=dates)
    >>> bet_timing(positions)
    DatetimeIndex(['2020-01-04', '2020-01-05', '2020-01-06'], \
dtype='datetime64[ns]', freq=None)
    """
    # 1. Bets at points where position returns to 0
    zero_positions = target_positions[target_positions == 0].index

    # 2. Bets at points where position flips sign
    # Find timestamps where t=0 and t-1 != 0
    lagged_non_zero = target_positions.shift(1)
    lagged_non_zero = lagged_non_zero[lagged_non_zero != 0].index
    bets = zero_positions.intersection(lagged_non_zero)

    # Find timestamps where sign flips, e.g., 1 to -1
    sign_flips = target_positions.iloc[1:] * target_positions.iloc[:-1].values
    bets = bets.union(sign_flips[sign_flips < 0].index).sort_values()

    # 3. Add last timestamp if not already included
    if target_positions.index[-1] not in bets:
        bets = bets.append(pd.Index([target_positions.index[-1]]))

    return bets

def calculate_holding_period(
    target_positions: pd.Series,
) -> Tuple[pd.DataFrame, float]:
    """
    Derive the average holding period in days.

    Uses the average entry time pairing algorithm.

    Parameters
    ----------
    target_positions : pd.Series
        Time series of target positions.

    Returns
    -------
    Tuple[pd.DataFrame, float]
        - hold_period_df: DataFrame with holding time ('dT') and
                          weight ('w') for each bet.
        - mean_holding_period: The weighted-average holding period in days.
    """
    hold_period_data = []
    time_entry = 0.0  # Average entry time
    position_diff = target_positions.diff()
    # Time difference in fractional days
    time_diff = (
        target_positions.index - target_positions.index[0]
    ) / np.timedelta64(1, "D")

    for i in range(1, target_positions.shape[0]):
        current_pos = target_positions.iloc[i]
        prev_pos = target_positions.iloc[i - 1]
        diff = position_diff.iloc[i]

        if diff * prev_pos >= 0:  # Position increase or flat
            if current_pos != 0:
                # Update average entry time
                time_entry = (
                    time_entry * prev_pos + time_diff[i] * diff
                ) / current_pos
        else:  # Position decrease or flip
            if current_pos * prev_pos < 0:  # Position flip
                # Close old position
                hold_period_data.append(
                    {
                        "index": target_positions.index[i],
                        "dT": time_diff[i] - time_entry,
                        "w": abs(prev_pos),
                    }
                )
                time_entry = time_diff[i]  # Reset entry time
            else:  # Position decrease (but not flip)
                hold_period_data.append(
                    {
                        "index": target_positions.index[i],
                        "dT": time_diff[i] - time_entry,
                        "w": abs(diff),
                    }
                )
    
    if not hold_period_data:
        return pd.DataFrame(columns=['dT', 'w']), np.nan
        
    hold_period_df = pd.DataFrame(hold_period_data).set_index('index')
    
    if hold_period_df["w"].sum() > 0:
        mean_holding_period = (
            (hold_period_df["dT"] * hold_period_df["w"]).sum()
            / hold_period_df["w"].sum()
        )
    else:
        mean_holding_period = np.nan

    return hold_period_df, mean_holding_period

def calculate_hhi(bet_returns: pd.Series) -> float:
    """
    Calculate the Herfindahl-Hirschman Index (HHI) for concentration.

    HHI is normalized to be between 0 and 1.
    - 0 indicates a perfectly diversified portfolio.
    - 1 indicates a fully concentrated portfolio.

    Parameters
    ----------
    bet_returns : pd.Series
        Series of returns for each bet.

    Returns
    -------
    float
        The normalized HHI value, or np.nan if undefined.
    """
    if bet_returns.shape[0] <= 2:
        return np.nan

    total_return = bet_returns.sum()
    if total_return == 0:
        return np.nan

    weights = bet_returns / total_return
    hhi = (weights**2).sum()
    
    # Normalize HHI
    n = bet_returns.shape[0]
    hhi_normalized = (hhi - 1.0 / n) / (1.0 - 1.0 / n)

    return hhi_normalized

def calculate_hhi_concentration(returns: pd.Series) -> Tuple[float, float, float]:
    """
    Calculate HHI concentration for positive, negative, and monthly returns.

    Parameters
    ----------
    returns : pd.Series
        Time series of returns.

    Returns
    -------
    Tuple[float, float, float]
        - hhi_positive: HHI for positive returns.
        - hhi_negative: HHI for negative returns.
        - hhi_time: HHI for time-based (monthly) concentration.
    """
    hhi_positive = calculate_hhi(returns[returns >= 0])
    hhi_negative = calculate_hhi(returns[returns < 0])
    
    # Calculate time concentration (by month)
    time_concentration = returns.groupby(pd.Grouper(freq="M")).count()
    hhi_time = calculate_hhi(time_concentration)

    return hhi_positive, hhi_negative, hhi_time

def compute_drawdowns_time_under_water(
    pnl_series: pd.Series, dollars: bool = False
) -> Tuple[pd.Series, pd.Series]:
    """
    Compute series of drawdowns and the time under water.

    Parameters
    ----------
    pnl_series : pd.Series
        Time series of returns or dollar PnL.
    dollars : bool, default=False
        If True, treat `pnl_series` as dollar PnL.
        If False, treat as returns (e.g., 1.01, 0.99) and
        calculate drawdowns as percentages.

    Returns
    -------
    Tuple[pd.Series, pd.Series]
        - drawdown: pd.Series indexed by start date, showing max drawdown.
        - time_under_water: pd.Series indexed by start date, showing
                            duration of drawdown in fractional years.
    """
    series_df = pnl_series.to_frame("PnL")
    series_df["HWM"] = series_df["PnL"].expanding().max()

    # Group by High Water Mark (HWM)
    groups = series_df.groupby("HWM")
    drawdown_analysis_data = []

    for hwm, group in groups:
        if len(group) <= 1 or hwm == group["PnL"].min():
            continue  # Skip groups with no drawdown

        drawdown_analysis_data.append(
            {
                "Start": group.index[0],
                "Stop": group.index[-1],
                "HWM": hwm,
                "Min": group["PnL"].min(),
            }
        )

    if not drawdown_analysis_data:
        return pd.Series(dtype=float), pd.Series(dtype=float)

    drawdown_analysis_df = pd.DataFrame(drawdown_analysis_data)
    drawdown_analysis_df = drawdown_analysis_df.set_index("Start")
    
    if dollars:
        drawdown = drawdown_analysis_df["HWM"] - drawdown_analysis_df["Min"]
    else:
        # Percentage drawdown
        drawdown = 1.0 - (drawdown_analysis_df["Min"] / drawdown_analysis_df["HWM"])

    # Time under water in fractional years
    time_under_water = (
        drawdown_analysis_df["Stop"] - drawdown_analysis_df.index
    ) / np.timedelta64(1, "Y")
    
    drawdown.index.name = 'Datetime'
    time_under_water.index.name = 'Datetime'

    return drawdown, time_under_water


"""
Generates synthetic price data using a Heston-Merton model
with Markov-switching regimes.
"""

from typing import Dict, List, Tuple, Union, Optional

import numpy as np
import pandas as pd
import quantecon.markov as qe
from numba import jit
from joblib import Parallel, delayed

# Type hint for regime parameters
RegimeParams = Dict[str, Union[float, List[float]]]

@jit(nopython=True)
def compute_log_returns(
    n_steps: int,
    mu_vector: np.ndarray,
    kappa_vector: np.ndarray,
    theta_vector: np.ndarray,
    xi_vector: np.ndarray,
    dw_stock: np.ndarray,
    dw_vol: np.ndarray,
    jump_comp: np.ndarray,
    poisson_var: np.ndarray,
    dt: float,
    sqrt_dt: float,
    lambda_vector: np.ndarray,
    m_vector: np.ndarray,
    v_vector: np.ndarray,
    regime_change: np.ndarray,
) -> np.ndarray:
    r"""
    Compute log returns using the Heston-Merton model (Numba-optimized).

    Heston Model (Volatility):
    .. math::
       v_{i+1} = v_i + \kappa_i (\theta_i - \max(v_i, 0)) dt +
                 \xi_i \sqrt{\max(v_i, 0)} dwV_i \sqrt{dt}

    Merton Jump-Diffusion (Price):
    .. math::
       \log(S_{i+1}/S_i) = \left(\mu_i - 0.5 v_i - \lambda_i (m_i +
                          \frac{v^2_i}{2})\right) dt +
                          \sqrt{v_i} dwS_i \sqrt{dt} + dJ_i

    Parameters
    ----------
    n_steps : int
        Number of steps (N).
    mu_vector : np.ndarray
        Drift vector.
    kappa_vector : np.ndarray
        Mean-reversion speed vector.
    theta_vector : np.ndarray
        Long-term mean vector for volatility.
    xi_vector : np.ndarray
        Volatility of volatility vector.
    dw_stock : np.ndarray
        Wiener process for stock.
    dw_vol : np.ndarray
        Wiener process for volatility.
    jump_comp : np.ndarray
        Jump component (Y).
    poisson_var : np.ndarray
        Poisson random variable vector (n).
    dt : float
        Time step.
    sqrt_dt : float
        Square root of the time step.
    lambda_vector : np.ndarray
        Jump intensity vector.
    m_vector : np.ndarray
        Mean of jump size vector.
    v_vector : np.ndarray
        Variance of jump size vector.
    regime_change : np.ndarray
        Boolean array, True if regime changed at this step.

    Returns
    -------
    np.ndarray
        Array of log returns.
    """
    v = np.empty(n_steps + 1)
    jump_events = np.multiply(poisson_var, jump_comp)
    v[0] = theta_vector[0]  # Initial volatility
    log_returns = np.empty(n_steps)

    for i in range(n_steps):
        if regime_change[i]:
            v[i] = theta_vector[i]  # Reset vol to new regime's long-term mean

        # Ensure v[i] is non-negative for sqrt
        v_i_safe = max(v[i], 0.0)
        
        # Volatility process (Heston)
        v[i + 1] = (
            v[i]
            + kappa_vector[i] * (theta_vector[i] - v_i_safe) * dt
            + xi_vector[i] * np.sqrt(v_i_safe) * dw_vol[i] * sqrt_dt
        )

        # Log return (Merton Jump-Diffusion)
        log_returns[i] = (
            (
                mu_vector[i]
                - 0.5 * v_i_safe
                - lambda_vector[i] * (m_vector[i] + (v_vector[i] ** 2) / 2.0)
            )
            * dt
            + np.sqrt(v_i_safe) * dw_stock[i] * sqrt_dt
            + jump_events[i]
        )

    return log_returns


def heston_merton_log_returns(
    total_time: float,
    n_steps: int,
    mu_vector: np.ndarray,
    kappa_vector: np.ndarray,
    theta_vector: np.ndarray,
    xi_vector: np.ndarray,
    rho_vector: np.ndarray,
    lambda_vector: np.ndarray,
    m_vector: np.ndarray,
    v_vector: np.ndarray,
    regime_change: np.ndarray,
    random_state: Optional[int] = None,
) -> np.ndarray:
    """
    Generate Heston-Merton log returns.

    This function sets up the correlated random numbers and calls
    the Numba-jitted `compute_log_returns`.

    Parameters
    ----------
    total_time : float
        Total time (T).
    n_steps : int
        Number of steps (N).
    *..._vector : np.ndarray
        Arrays of model parameters (see `compute_log_returns`).
    regime_change : np.ndarray
        Boolean array, True if regime changed at this step.
    random_state : int, optional
        Seed for the random number generator.

    Returns
    -------
    np.ndarray
        Array of log returns.
    """
    # Ensure vectors are of length n_steps
    params = [
        mu_vector, kappa_vector, theta_vector, xi_vector, rho_vector,
        lambda_vector, m_vector, v_vector
    ]
    if not all(len(p) == n_steps for p in params):
        raise ValueError("All parameter vectors must have length `n_steps`")
    
    rng = np.random.default_rng(random_state)
    dt = total_time / n_steps
    sqrt_dt = np.sqrt(dt)

    # Generate correlated random numbers
    z = np.zeros((n_steps, 3))
    n = np.zeros(n_steps)
    for i in range(n_steps):
        cov_matrix = np.array([
            [1.0,           rho_vector[i], 0.0],
            [rho_vector[i], 1.0,           0.0],
            [0.0,           0.0,           v_vector[i] ** 2],
        ])
        z[i] = rng.multivariate_normal(
            [0.0, 0.0, m_vector[i]], cov_matrix
        )
        n[i] = rng.poisson(lambda_vector[i] * dt)

    dw_stock = z[:, 0]
    dw_vol = z[:, 1]
    jump_comp = z[:, 2]

    return compute_log_returns(
        n_steps,
        mu_vector,
        kappa_vector,
        theta_vector,
        xi_vector,
        dw_stock,
        dw_vol,
        jump_comp,
        n,
        dt,
        sqrt_dt,
        lambda_vector,
        m_vector,
        v_vector,
        regime_change,
    )


def align_params_length(
    regime_params: RegimeParams,
) -> Tuple[Dict[str, List[float]], int]:
    """
    Align the parameter lists within a regime to be the same length.

    If a parameter is a float, it's broadcast to a list.
    If a parameter is a list shorter than the max, it's extended
    by repeating its last value.

    Parameters
    ----------
    regime_params : RegimeParams
        Dictionary of regime parameters.

    Returns
    -------
    Tuple[Dict[str, List[float]], int]
        - The aligned regime parameter dictionary.
        - The maximum length (number of steps) for this regime.
    """
    max_len = max(
        len(v) if isinstance(v, list) else 1 for v in regime_params.values()
    )

    aligned_params: Dict[str, List[float]] = {}
    for key, value in regime_params.items():
        if isinstance(value, list):
            if len(value) < max_len:
                # Extend list by repeating last value
                aligned_params[key] = value + [value[-1]] * (max_len - len(value))
            else:
                aligned_params[key] = value[:max_len] # Truncate if too long
        else:
            # Broadcast float to list
            aligned_params[key] = [value] * max_len

    return aligned_params, max_len


def generate_prices_from_regimes(
    regimes: Dict[str, RegimeParams],
    transition_matrix: np.ndarray,
    total_time: float,
    n_steps: int,
    random_state: Optional[int] = None,
) -> Tuple[pd.Series, np.ndarray]:
    """
    Generate a price series from a Markov-switching regime model.

    Parameters
    ----------
    regimes : Dict[str, RegimeParams]
        Dictionary mapping regime names to their parameter sets.
    transition_matrix : np.ndarray
        The Markov Chain transition matrix.
    total_time : float
        Total time for the simulation (e.g., 1.0 for 1 year).
    n_steps : int
        Number of discrete steps in the simulation.
    random_state : int, optional
        Seed for random number generation.

    Returns
    -------
    Tuple[pd.Series, np.ndarray]
        - The generated price series, indexed by business days.
        - The array of simulated regime names for each step.
    """
    rng = np.random.default_rng(random_state)
    
    # 1. Simulate the Markov Chain
    regime_names = list(regimes.keys())
    markov_chain = qe.MarkovChain(transition_matrix, state_values=regime_names)
    simulated_regimes = markov_chain.simulate(
        ts_length=n_steps, random_state=rng
    )

    # 2. Unpack parameters based on simulated regimes
    param_lists: Dict[str, List[float]] = {
        "mu": [], "kappa": [], "theta": [], "xi": [],
        "rho": [], "lam": [], "m": [], "v": [],
    }
    
    regime_path_expanded = []
    
    current_step = 0
    while current_step < n_steps:
        regime_name = simulated_regimes[current_step]
        params, regime_len = align_params_length(regimes[regime_name].copy())
        
        steps_to_take = min(regime_len, n_steps - current_step)
        
        for key in param_lists:
            param_lists[key].extend(params[key][:steps_to_take])
            
        regime_path_expanded.extend([regime_name] * steps_to_take)
        current_step += steps_to_take

    # 3. Finalize parameter arrays and regime path
    simulated_regimes_final = np.array(regime_path_expanded)
    param_arrays: Dict[str, np.ndarray] = {
        key: np.array(val) for key, val in param_lists.items()
    }

    # 4. Identify regime change points
    regime_change = simulated_regimes_final[1:] != simulated_regimes_final[:-1]
    regime_change = np.pad(
        regime_change, (1, 0), mode="constant", constant_values=False
    )

    # 5. Generate log returns
    log_returns = heston_merton_log_returns(
        total_time,
        n_steps,
        param_arrays["mu"],
        param_arrays["kappa"],
        param_arrays["theta"],
        param_arrays["xi"],
        param_arrays["rho"],
        param_arrays["lam"],
        param_arrays["m"],
        param_arrays["v"],
        regime_change,
        random_state=rng.integers(0, 1e6),
    )

    # 6. Create price series with a Business Day index
    start_day = "2000-01-01"
    business_days = pd.date_range(
        start=start_day, periods=n_steps, freq="B"
    )
    
    price_series = pd.Series(log_returns, index=business_days).ffill()
    prices = 100 * np.exp(price_series.cumsum())
    prices.name = "Price"

    return prices, simulated_regimes_final


def parallel_generate_prices(
    number_of_paths: int,
    regimes: Dict[str, RegimeParams],
    transition_matrix: np.ndarray,
    total_time: float,
    number_of_steps: int,
    random_state: Optional[int] = None,
    n_jobs: int = 1,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Parallel generation of price paths.

    Parameters
    ----------
    number_of_paths : int
        The number of paths to generate.
    regimes : Dict[str, RegimeParams]
        Dictionary of regime parameters.
    transition_matrix : np.ndarray
        Markov Chain transition matrix.
    total_time : float
        Total time for the simulation.
    number_of_steps : int
        Number of steps in the simulation.
    random_state : int, optional
        Seed for *main* random number generator (which generates seeds).
    n_jobs : int, default=1
        Number of parallel jobs to run.

    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame]
        - prices_df: DataFrame of (n_steps x number_of_paths) prices.
        - regimes_df: DataFrame of (n_steps x number_of_paths) regimes.
    """
    rng = np.random.default_rng(random_state)
    # Generate unique seeds for each parallel job
    random_states = rng.integers(0, 10 * number_of_paths, size=number_of_paths)
    
    results = Parallel(n_jobs=n_jobs)(
        delayed(generate_prices_from_regimes)(
            regimes,
            transition_matrix,
            total_time,
            number_of_steps,
            seed,
        )
        for seed in random_states
    )

    prices, simulated_regimes = zip(*results)

    prices_df = pd.concat(prices, axis=1)
    prices_df.columns = range(number_of_paths)
    
    simulated_regimes_df = pd.DataFrame(simulated_regimes).T
    simulated_regimes_df.columns = range(number_of_paths)
    simulated_regimes_df.index = prices_df.index

    return prices_df, simulated_regimes_df


"""
Implements the Probability Mass Function (PMF) calculation.
"""

from collections import Counter
from typing import Dict

def probability_mass_function(
    message: str, approximate_word_length: int
) -> Dict[str, float]:
    """
    Calculate the Probability Mass Function (PMF) of n-grams.

    Parameters
    ----------
    message : str
        Input string.
    approximate_word_length : int
        The length of the "words" or n-grams to analyze (e.g., 1, 2, 3).

    Returns
    -------
    Dict[str, float]
        A dictionary mapping each n-gram to its probability.
    """
    if not message or len(message) < approximate_word_length:
        return {}
        
    # Find all n-grams (words)
    library = Counter(
        message[i : i + approximate_word_length]
        for i in range(len(message) - approximate_word_length + 1)
    )

    # The total number of n-grams
    num_windows = float(len(message) - approximate_word_length + 1)
    if num_windows == 0:
        return {}

    # Calculate probability for each n-gram
    pmf = {
        key: count / num_windows for key, count in library.items()
    }

    return pmf


"""
Implements the core logic for the (G)SADF (Generalized) Supreme
Augmented Dickey-Fuller test for structural breaks.

Reference:
    De Prado, M. (2018) Advances in financial machine learning.
    John Wiley & Sons, Chapter 17.
"""

import numpy as np
import pandas as pd
from typing import List, Union, Tuple, Dict, Any

def lag_dataframe(
    market_data: pd.DataFrame, lags: Union[int, List[int]]
) -> pd.DataFrame:
    """
    Apply lags to a DataFrame.

    Reference:
        Snippet 17.3, Page 253.

    Parameters
    ----------
    market_data : pd.DataFrame
        DataFrame of price or log price.
    lags : Union[int, List[int]]
        An integer number of lags (e.g., 3 creates lags 0, 1, 2, 3)
        or a specific list of lags to create.

    Returns
    -------
    pd.DataFrame
        A DataFrame with lagged columns, e.g., 'price_0', 'price_1', ...
    """
    lagged_parts = []
    
    if isinstance(lags, int):
        lags_list = range(lags + 1)
    else:
        lags_list = [int(lag) for lag in lags]

    for lag in lags_list:
        temp_data = market_data.shift(lag).copy()
        temp_data.columns = [f"{col}_{lag}" for col in temp_data.columns]
        lagged_parts.append(temp_data)

    return pd.concat(lagged_parts, axis=1)


def prepare_data(
    series: pd.DataFrame, constant: str, lags: int
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Prepare the y and X matrices for ADF regression.

    Reference:
        Snippet 17.2, Page 252.

    Parameters
    ----------
    series : pd.DataFrame
        A single-column DataFrame of price or log price.
    constant : str
        The type of trend to include:
        - 'nc': No constant or trend.
        - 'c': Constant only.
        - 'ct': Constant and linear trend.
        - 'ctt': Constant, linear, and quadratic trend.
    lags : int
        The number of lags to include in the regression.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        - y: The dependent variable (differenced series).
        - X: The matrix of independent variables (lagged level,
             lagged diffs, and trend components).
    """
    # 1. Create lagged differenced series
    series_diff = series.diff().dropna()
    x = lag_dataframe(series_diff, lags).dropna()
    
    # 2. Replace lag 0 of diffs with lag 1 of levels
    x.iloc[:, 0] = series.values[-x.shape[0] - 1 : -1, 0]
    y = series_diff.iloc[-x.shape[0] :].values

    x_arr = x.to_numpy()

    # 3. Add trend components
    if constant != "nc":
        # Add constant
        x_arr = np.append(x_arr, np.ones((x_arr.shape[0], 1)), axis=1)

    if constant[:2] == "ct":
        # Add linear trend
        trend = np.arange(x_arr.shape[0]).reshape(-1, 1)
        x_arr = np.append(x_arr, trend, axis=1)

    if constant == "ctt":
        # Add quadratic trend
        x_arr = np.append(x_arr, trend**2, axis=1)

    return y, x_arr


def compute_beta(
    y: np.ndarray, x: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Fit the ADF specification using OLS.

    Reference:
        Snippet 17.4, Page 253.

    Parameters
    ----------
    y : np.ndarray
        Dependent variable.
    x : np.ndarray
        Matrix of independent variables.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        - beta_mean: The OLS coefficient estimates.
        - beta_variance: The OLS covariance matrix.
    """
    # (X'X)^-1
    x_inv = np.linalg.inv(x.T @ x)
    
    # beta = (X'X)^-1 @ (X'y)
    beta_mean = x_inv @ (x.T @ y)
    
    # epsilon = y - X @ beta
    epsilon = y - (x @ beta_mean)
    
    # V[beta] = (e'e / (T-K)) * (X'X)^-1
    beta_variance = (
        (epsilon.T @ epsilon) / (x.shape[0] - x.shape[1])
    ) * x_inv

    return beta_mean, beta_variance


def adf(
    log_price: pd.DataFrame,
    min_sample_length: int,
    constant: str,
    lags: int,
) -> Dict[str, Any]:
    """
    Run the ADF test over expanding windows (SADF's inner loop).

    This function computes the ADF statistic for all possible
    start dates, finding the maximum (Supremum) ADF statistic.

    Reference:
        Snippet 17.1, Page 251.

    Parameters
    ----------
    log_price : pd.DataFrame
        DataFrame of log prices.
    min_sample_length : int
        Minimum sample length for the regression.
    constant : str
        Trend component ('nc', 'c', 'ct', 'ctt').
    lags : int
        Number of lags to include.

    Returns
    -------
    Dict[str, Any]
        A dictionary containing:
        - 'Time': The timestamp of the end of the series.
        - 'gsadf': The BSADF statistic (the supremum ADF).
    """
    # 1. Prepare the full X, y matrices
    y, x = prepare_data(log_price, constant=constant, lags=lags)
    
    # 2. Define all possible start points
    start_points = range(0, y.shape[0] + lags - min_sample_length + 1)
    bsadf = -np.inf  # Supremum ADF
    all_adf = []

    # 3. Loop over all expanding windows
    for start in start_points:
        y_window, x_window = y[start:], x[start:]
        
        # 4. Compute ADF regression for this window
        beta_mean, beta_variance = compute_beta(y_window, x_window)
        
        # 5. Get t-statistic for the first coefficient (the level)
        beta_mean_level = beta_mean[0, 0]
        beta_std_level = beta_variance[0, 0] ** 0.5
        
        if beta_std_level == 0:
            t_stat = -np.inf # Avoid division by zero
        else:
            t_stat = beta_mean_level / beta_std_level
            
        all_adf.append(t_stat)

        if t_stat > bsadf:
            bsadf = t_stat # Update supremum

    out = {"Time": log_price.index[-1], "gsadf": bsadf}
    return out


"""
RiskLabAI Structural Breaks Features Module

Implements methods for detecting structural breaks in time series,
including the Augmented Dickey-Fuller (ADF) and (G)SADF tests
as described by de Prado.
"""

from .structural_breaks import (
    lag_dataframe,
    prepare_data,
    compute_beta,
    adf,
)

__all__ = [
    "lag_dataframe",
    "prepare_data",
    "compute_beta",
    "adf",
]


"""
Controller class to act as a factory for initializing
various bar types (Standard, Time, Imbalance, Run).
"""

from typing import Tuple, Union, Dict, Callable, Optional, Any
import pandas as pd

# Import bar types
from RiskLabAI.data.structures.imbalance_bars import (
    ExpectedImbalanceBars, FixedImbalanceBars
)
from RiskLabAI.data.structures.run_bars import ExpectedRunBars, FixedRunBars
from RiskLabAI.data.structures.standard_bars import StandardBars
from RiskLabAI.data.structures.time_bars import TimeBars
from RiskLabAI.data.structures.abstract_bars import AbstractBars

# Import constants
from RiskLabAI.utils.constants import (
    CUMULATIVE_DOLLAR, CUMULATIVE_VOLUME, CUMULATIVE_TICKS
)

class BarsInitializerController:
    """
    Controller for initializing various types of bars.

    This class acts as a factory, providing static methods to
    construct different bar objects with sensible default parameters.
    """

    def __init__(self):
        """
        Initializes the controller and maps method names to methods.
        """
        self.method_name_to_method: Dict[str, Callable[..., AbstractBars]] = {
            "expected_dollar_imbalance_bars": self.initialize_expected_dollar_imbalance_bars,
            "expected_volume_imbalance_bars": self.initialize_expected_volume_imbalance_bars,
            "expected_tick_imbalance_bars": self.initialize_expected_tick_imbalance_bars,
            "fixed_dollar_imbalance_bars": self.initialize_fixed_dollar_imbalance_bars,
            "fixed_volume_imbalance_bars": self.initialize_fixed_volume_imbalance_bars,
            "fixed_tick_imbalance_bars": self.initialize_fixed_tick_imbalance_bars,
            "expected_dollar_run_bars": self.initialize_expected_dollar_run_bars,
            "expected_volume_run_bars": self.initialize_expected_volume_run_bars,
            "expected_tick_run_bars": self.initialize_expected_tick_run_bars,
            "fixed_dollar_run_bars": self.initialize_fixed_dollar_run_bars,
            "fixed_volume_run_bars": self.initialize_fixed_volume_run_bars,
            "fixed_tick_run_bars": self.initialize_fixed_tick_run_bars,
            "dollar_standard_bars": self.initialize_dollar_standard_bars,
            "volume_standard_bars": self.initialize_volume_standard_bars,
            "tick_standard_bars": self.initialize_tick_standard_bars,
            "time_bars": self.initialize_time_bars,
        }

    @staticmethod
    def initialize_expected_dollar_imbalance_bars(
        window_size_for_expected_n_ticks_estimation: int = 3,
        window_size_for_expected_imbalance_estimation: int = 10000,
        initial_estimate_of_expected_n_ticks_in_bar: int = 20000,
        expected_ticks_number_bounds: Optional[Tuple[float, float]] = None,
        analyze_thresholds: bool = False,
    ) -> ExpectedImbalanceBars:
        """
        Initialize expected dollar imbalance bars.

        Parameters
        ----------
        window_size_for_expected_n_ticks_estimation : int
            Window size for E[T] estimation.
        window_size_for_expected_imbalance_estimation : int
            Window size for E[theta] estimation.
        initial_estimate_of_expected_n_ticks_in_bar : int
            Initial E[T] guess.
        expected_ticks_number_bounds : Tuple[float, float], optional
            Bounds for E[T].
        analyze_thresholds : bool
            Flag to analyze thresholds.

        Returns
        -------
        ExpectedImbalanceBars
            An instance configured for dollar imbalance.
        """
        return ExpectedImbalanceBars(
            bar_type="dollar_imbalance",
            window_size_for_expected_n_ticks_estimation=window_size_for_expected_n_ticks_estimation,
            initial_estimate_of_expected_n_ticks_in_bar=initial_estimate_of_expected_n_ticks_in_bar,
            window_size_for_expected_imbalance_estimation=window_size_for_expected_imbalance_estimation,
            expected_ticks_number_bounds=expected_ticks_number_bounds,
            analyse_thresholds=analyze_thresholds,
        )

    @staticmethod
    def initialize_expected_volume_imbalance_bars(
        **kwargs: Any,
    ) -> ExpectedImbalanceBars:
        """Initialize expected volume imbalance bars."""
        return ExpectedImbalanceBars(bar_type="volume_imbalance", **kwargs)

    @staticmethod
    def initialize_expected_tick_imbalance_bars(
        **kwargs: Any,
    ) -> ExpectedImbalanceBars:
        """Initialize expected tick imbalance bars."""
        return ExpectedImbalanceBars(bar_type="tick_imbalance", **kwargs)

    @staticmethod
    def initialize_fixed_dollar_imbalance_bars(
        **kwargs: Any,
    ) -> FixedImbalanceBars:
        """Initialize fixed dollar imbalance bars."""
        return FixedImbalanceBars(bar_type="dollar_imbalance", **kwargs)

    @staticmethod
    def initialize_fixed_volume_imbalance_bars(
        **kwargs: Any,
    ) -> FixedImbalanceBars:
        """Initialize fixed volume imbalance bars."""
        return FixedImbalanceBars(bar_type="volume_imbalance", **kwargs)

    @staticmethod
    def initialize_fixed_tick_imbalance_bars(
        **kwargs: Any,
    ) -> FixedImbalanceBars:
        """Initialize fixed tick imbalance bars."""
        return FixedImbalanceBars(bar_type="tick_imbalance", **kwargs)

    @staticmethod
    def initialize_expected_dollar_run_bars(
        **kwargs: Any,
    ) -> ExpectedRunBars:
        """Initialize expected dollar run bars."""
        return ExpectedRunBars(bar_type="dollar_run", **kwargs)

    @staticmethod
    def initialize_expected_volume_run_bars(
        **kwargs: Any,
    ) -> ExpectedRunBars:
        """Initialize expected volume run bars."""
        return ExpectedRunBars(bar_type="volume_run", **kwargs)

    @staticmethod
    def initialize_expected_tick_run_bars(
        **kwargs: Any,
    ) -> ExpectedRunBars:
        """Initialize expected tick run bars."""
        return ExpectedRunBars(bar_type="tick_run", **kwargs)

    @staticmethod
    def initialize_fixed_dollar_run_bars(
        **kwargs: Any,
    ) -> FixedRunBars:
        """Initialize fixed dollar run bars."""
        return FixedRunBars(bar_type="dollar_run", **kwargs)

    @staticmethod
    def initialize_fixed_volume_run_bars(
        **kwargs: Any,
    ) -> FixedRunBars:
        """Initialize fixed volume run bars."""
        return FixedRunBars(bar_type="volume_run", **kwargs)

    @staticmethod
    def initialize_fixed_tick_run_bars(
        **kwargs: Any,
    ) -> FixedRunBars:
        """Initialize fixed tick run bars."""
        return FixedRunBars(bar_type="tick_run", **kwargs)

    @staticmethod
    def initialize_dollar_standard_bars(
        threshold: Union[float, pd.Series] = 70000000,
    ) -> StandardBars:
        """Initialize dollar standard bars."""
        return StandardBars(bar_type=CUMULATIVE_DOLLAR, threshold=threshold)

    @staticmethod
    def initialize_volume_standard_bars(
        threshold: Union[float, pd.Series] = 30000,
    ) -> StandardBars:
        """Initialize volume standard bars."""
        return StandardBars(bar_type=CUMULATIVE_VOLUME, threshold=threshold)

    @staticmethod
    def initialize_tick_standard_bars(
        threshold: Union[float, pd.Series] = 6000,
    ) -> StandardBars:
        """Initialize tick standard bars."""
        return StandardBars(bar_type=CUMULATIVE_TICKS, threshold=threshold)

    @staticmethod
    def initialize_time_bars(
        resolution_type: str = "D",
        resolution_units: int = 1,
    ) -> TimeBars:
        """Initialize time bars."""
        return TimeBars(
            resolution_type=resolution_type, resolution_units=resolution_units
        )


"""
Implements the Shannon Entropy estimator.
"""

from collections import Counter
from math import log2

def shannon_entropy(message: str) -> float:
    """
    Calculate the Shannon Entropy of a message.

    Parameters
    ----------
    message : str
        Input string (e.g., a discretized time series "1-1101...").

    Returns
    -------
    float
        The calculated Shannon Entropy (in bits).
    """
    if not message:
        return 0.0

    character_counts = Counter(message)
    message_length = len(message)

    entropy = -sum(
        (count / message_length) * log2(count / message_length)
        for count in character_counts.values()
    )

    return entropy


"""
Implements the core financial labeling functions, including:
- CUSUM filters for event sampling.
- Volatility estimation.
- The Triple-Barrier Method (vertical and horizontal barriers).
- Meta-labeling.
- Multiprocessing helpers for parallel execution.

Reference:
    De Prado, M. (2018) Advances in financial machine learning.
    John Wiley & Sons, Chapters 3 & 4.
"""

import sys
import time
import datetime
import multiprocessing as mp
from concurrent.futures import ProcessPoolExecutor
from typing import List, Optional, Tuple, Callable, Dict, Any

import numpy as np
import pandas as pd


def cusum_filter_events_dynamic_threshold(
    prices: pd.Series, threshold: pd.Series
) -> pd.DatetimeIndex:
    """
    Detect events using the Symmetric CUSUM filter with a dynamic threshold.

    This filter identifies timestamps where the cumulative sum of price
    changes exceeds a dynamic, time-varying threshold.

    Reference:
        Snippet 3.2, Page 48 (modified for dynamic threshold).

    Parameters
    ----------
    prices : pd.Series
        A pandas Series of prices.
    threshold : pd.Series
        A pandas Series containing the threshold values for each timestamp.
        Must be aligned with the `prices` index.

    Returns
    -------
    pd.DatetimeIndex
        Timestamps of the detected events (when a barrier was touched).
    """
    time_events = []
    shift_positive, shift_negative = 0.0, 0.0
    price_delta = prices.diff().dropna()
    
    # Align price changes with thresholds
    price_delta, thresholds = price_delta.align(
        threshold, join="inner", copy=False
    )

    for (index, value), thresh_val in zip(
        price_delta.items(), thresholds.values
    ):
        shift_positive = max(0.0, shift_positive + value)
        shift_negative = min(0.0, shift_negative + value)

        if shift_negative < -thresh_val:
            shift_negative = 0.0
            time_events.append(index)
        elif shift_positive > thresh_val:
            shift_positive = 0.0
            time_events.append(index)

    return pd.DatetimeIndex(time_events)


def symmetric_cusum_filter(
    prices: pd.Series, threshold: float
) -> pd.DatetimeIndex:
    """
    Detect events using the Symmetric CUSUM filter with a fixed threshold.

    Reference:
        Snippet 3.2, Page 48.

    Parameters
    ----------
    prices : pd.Series
        A pandas Series of prices.
    threshold : float
        The fixed threshold value.

    Returns
    -------
    pd.DatetimeIndex
        Timestamps of the detected events (when a barrier was touched).
    """
    time_events = []
    shift_positive, shift_negative = 0.0, 0.0
    price_delta = prices.diff().dropna()

    for index, value in price_delta.items():
        shift_positive = max(0.0, shift_positive + value)
        shift_negative = min(0.0, shift_negative + value)

        if shift_negative < -threshold:
            shift_negative = 0.0
            time_events.append(index)
        elif shift_positive > threshold:
            shift_positive = 0.0
            time_events.append(index)

    return pd.DatetimeIndex(time_events)


def daily_volatility_with_log_returns(
    close: pd.Series, span: int = 100
) -> pd.Series:
    """
    Calculate daily volatility using log returns.

    This method computes the EWMA standard deviation of daily log returns.

    Reference:
        Snippet 3.1, Page 44.

    Parameters
    ----------
    close : pd.Series
        Time series of close prices.
    span : int, default=100
        The span parameter for the EWMA.

    Returns
    -------
    pd.Series
        Series of daily volatility estimates.
    """
    # Find timestamps of 1 day prior
    df = close.index.searchsorted(close.index - pd.Timedelta(days=1))
    df = df[df > 0]
    df = pd.Series(
        close.index[df - 1],
        index=close.index[close.shape[0] - df.shape[0] :],
    )
    
    # Calculate log returns
    returns = np.log(close.loc[df.index] / close.loc[df.values].values)
    stds = returns.ewm(span=span).std().rename("std")
    return stds


def vertical_barrier(
    close: pd.Series, time_events: pd.DatetimeIndex, number_days: int
) -> pd.Series:
    """
    Create a vertical barrier for the triple-barrier method.

    This function finds the timestamp that occurs `number_days` after
    each event in `time_events`.

    Reference:
        Snippet 3.4, Page 50.

    Parameters
    ----------
    close : pd.Series
        Time series of close prices (used for the index).
    time_events : pd.DatetimeIndex
        The timestamps of the events (e.g., from CUSUM filter).
    number_days : int
        The number of days to look forward for the vertical barrier.

    Returns
    -------
    pd.Series
        A Series where the index is the event timestamp and the
        value is the vertical barrier timestamp.
    """
    barrier_dates = time_events + pd.Timedelta(days=number_days)
    
    # Find the integer location of the barrier dates in the price index
    timestamp_array = close.index.searchsorted(barrier_dates)
    
    # Filter out any indices that are out of bounds
    valid_indices = timestamp_array < close.shape[0]
    timestamp_array = timestamp_array[valid_indices]
    
    barrier_series = pd.Series(
        close.index[timestamp_array],
        index=time_events[valid_indices],
    )
    return barrier_series


def triple_barrier(
    close: pd.Series,
    events: pd.DataFrame,
    ptsl: List[float],
    molecule: List[pd.Timestamp],
) -> pd.DataFrame:
    """
    Apply the triple-barrier method for a subset of events.

    This is the core worker function for `meta_events`. It finds the
    first touch time of the upper, lower, or vertical barrier.

    Reference:
        Snippet 3.3, Page 50.

    Parameters
    ----------
    close : pd.Series
        Time series of close prices.
    events : pd.DataFrame
        DataFrame with event info. Must contain 'End Time' (vertical
        barrier), 'Base Width' (volatility), and 'Side' (if applicable).
    ptsl : List[float]
        A list of two floats: [profit_taking_multiplier, stop_loss_multiplier].
    molecule : List[pd.Timestamp]
        The subset of event timestamps this worker is responsible for.

    Returns
    -------
    pd.DataFrame
        A DataFrame with columns ['stop_loss', 'profit_taking']
        indexed by the event timestamp, indicating the first
        touch time for each horizontal barrier.
    """
    # Filter for events this worker owns
    events_filtered = events.loc[molecule]
    output = pd.DataFrame(index=events_filtered.index)
    output["End Time"] = events_filtered["End Time"] # Use original end time

    # 1. Set horizontal barriers
    if ptsl[0] > 0:
        profit_taking = ptsl[0] * events_filtered["Base Width"]
    else:
        profit_taking = pd.Series(np.inf, index=events_filtered.index)

    if ptsl[1] > 0:
        stop_loss = -ptsl[1] * events_filtered["Base Width"]
    else:
        stop_loss = pd.Series(-np.inf, index=events_filtered.index)

    # Get side if it exists, otherwise default to 1 (long)
    side = events_filtered.get("Side", pd.Series(1.0, index=events_filtered.index))

    # 2. Find first touch time
    for location, vertical_barrier_time in events_filtered["End Time"].fillna(close.index[-1]).items():
        # Path prices from event start to vertical barrier
        path_prices = close.loc[location:vertical_barrier_time]
        
        # Calculate path returns, adjusted by side
        path_returns = (
            np.log(path_prices / close[location]) * side.at[location]
        )

        output.loc[location, "stop_loss"] = path_returns[
            path_returns < stop_loss.at[location]
        ].index.min()
        
        output.loc[location, "profit_taking"] = path_returns[
            path_returns > profit_taking.at[location]
        ].index.min()

    # The 'End Time' column in output now holds the *first* barrier touched
    output["End Time"] = output.min(axis=1)
    return output.drop(columns=["stop_loss", "profit_taking"])


def meta_events(
    close: pd.Series,
    time_events: pd.DatetimeIndex,
    ptsl: List[float],
    target: pd.Series,
    return_min: float,
    num_threads: int,
    vertical_barrier_times: Optional[pd.Series] = None,
    side: Optional[pd.Series] = None,
) -> pd.DataFrame:
    """
    Generate events for meta-labeling using the triple-barrier method.

    This function sets up and executes the triple-barrier search in parallel.

    Reference:
        Snippet 3.6, Page 51.

    Parameters
    ----------
    close : pd.Series
        Time series of close prices.
    time_events : pd.DatetimeIndex
        Timestamps of the events (e.g., from CUSUM filter).
    ptsl : List[float]
        [profit_taking_multiplier, stop_loss_multiplier].
    target : pd.Series
        Series of volatility (or 'Base Width') for setting barrier width.
    return_min : float
        The minimum target volatility required to run a search.
    num_threads : int
        Number of parallel threads to use.
    vertical_barrier_times : pd.Series, optional
        A Series of vertical barrier timestamps. If None, barriers are not used.
    side : pd.Series, optional
        A Series of position sides (1 or -1). If None, assumes long-only.

    Returns
    -------
    pd.DataFrame
        The events DataFrame with columns:
        - 'End Time': The timestamp of the *first* barrier touch.
        - 'Base Width': The volatility used for the barriers.
        - 'Side': (Optional) The side of the bet.
    """
    # 1. Filter events by volatility
    target = target.reindex(time_events)
    target = target[target > return_min]
    if target.empty:
        return pd.DataFrame(columns=["End Time", "Base Width", "Side"])

    # 2. Set up vertical barrier
    if vertical_barrier_times is None:
        vertical_barrier_times = pd.Series(pd.NaT, index=target.index)
    else:
        vertical_barrier_times = vertical_barrier_times.reindex(target.index)

    # 3. Set up sides
    if side is None:
        side_series = pd.Series(1.0, index=target.index)
        ptsl_final = [ptsl[0], ptsl[0]] # Symmetric barriers
    else:
        side_series = side.reindex(target.index)
        ptsl_final = ptsl[:2] # Asymmetric barriers

    # 4. Create base events DataFrame
    events = pd.concat(
        {
            "End Time": vertical_barrier_times,
            "Base Width": target,
            "Side": side_series,
        },
        axis=1,
    ).dropna(subset=["Base Width"])
    
    # 5. Run in parallel
    molecule_subsets = np.array_split(events.index, num_threads)
    
    with ProcessPoolExecutor(max_workers=num_threads) as executor:
        results = list(executor.map(
            triple_barrier,
            [close] * num_threads,
            [events] * num_threads,
            [ptsl_final] * num_threads,
            molecule_subsets
        ))

    # Combine results and update End Time
    first_touch_times = pd.concat(results, axis=0)["End Time"]
    events["End Time"] = first_touch_times.reindex(events.index)

    if side is None:
        events = events.drop("Side", axis=1)
    
    return events


def meta_labeling(
    events: pd.DataFrame, close: pd.Series
) -> pd.DataFrame:
    """
    Calculate returns and assign binary labels for meta-labeling.

    Reference:
        Snippet 3.7, Page 52.

    Parameters
    ----------
    events : pd.DataFrame
        The events DataFrame from `meta_events`.
    close : pd.Series
        Time series of close prices.

    Returns
    -------
    pd.DataFrame
        DataFrame with columns:
        - 'End Time': Time of first barrier touch.
        - 'Return': The return of the trade.
        - 'Label': The binary label (0 or 1 if 'Side' is present, -1 or 1 if not).
        - 'Side': (Optional) The side of the bet.
    """
    events_filtered = events.dropna(subset=["End Time"])
    
    # Get all unique timestamps
    all_dates = events_filtered.index.union(
        events_filtered["End Time"].values
    ).drop_duplicates()
    
    # Reindex prices to only the required dates
    close_filtered = close.reindex(all_dates, method="bfill")

    out = pd.DataFrame(index=events_filtered.index)
    out["End Time"] = events_filtered["End Time"]
    
    # Calculate returns
    out["Return"] = (
        close_filtered.loc[events_filtered["End Time"].values].values
        / close_filtered.loc[events_filtered.index].values
        - 1.0
    )

    if "Side" in events_filtered:
        out["Return"] *= events_filtered["Side"]
        out["Side"] = events_filtered["Side"]

    # Assign labels
    out["Label"] = np.sign(out["Return"])
    
    if "Side" in events_filtered:
        # Meta-labeling: 1 if profitable, 0 if not
        out.loc[out["Return"] <= 0, "Label"] = 0.0

    return out


# --- Multiprocessing Helper Functions ---
# (These are generic utilities)

def lin_parts(num_atoms: int, num_threads: int) -> np.ndarray:
    """
    Create linear partitions for parallel processing.

    Parameters
    ----------
    num_atoms : int
        Total number of items to process.
    num_threads : int
        Number of threads (partitions) to create.

    Returns
    -------
    np.ndarray
        Array of partition indices.
    """
    parts = np.linspace(0, num_atoms, min(num_threads, num_atoms) + 1)
    parts = np.ceil(parts).astype(int)
    return parts


def process_jobs(
    jobs: List[Dict[str, Any]], task: Optional[str] = None, num_threads: int = 24
) -> List[Any]:
    """
    Run jobs in parallel using multiprocessing.

    Parameters
    ----------
    jobs : List[Dict]
        List of job dictionaries. Each dict must contain a 'func' key
        and its corresponding arguments.
    task : str, optional
        Name of the task for progress reporting.
    num_threads : int, default=24
        Number of threads to use.

    Returns
    -------
    List[Any]
        A list containing the results from each job.
    """
    if task is None:
        task = jobs[0]["func"].__name__
    
    with mp.Pool(processes=num_threads) as pool:
        outputs = pool.imap_unordered(expand_call, jobs)
        out = []
        time0 = time.time()
        
        # Process async output, report progress
        for i, out_ in enumerate(outputs, 1):
            out.append(out_)
            report_progress(i, len(jobs), time0, task)
            
    return out


def expand_call(kargs: Dict[str, Any]) -> Any:
    """
    Worker function to expand keyword arguments and call the function.

    Parameters
    ----------
    kargs : Dict
        Dictionary of arguments, including the 'func' to be called.

    Returns
    -------
    Any
        The result of the function call.
    """
    func = kargs.pop('func')
    return func(**kargs)


def report_progress(
    job_num: int, num_jobs: int, time0: float, task: str
) -> None:
    """Report progress of parallel jobs."""
    msg = [float(job_num) / num_jobs, (time.time() - time0) / 60.0]
    msg.append(msg[1] * (1 / msg[0] - 1))  # Remaining time
    timestamp = str(datetime.datetime.fromtimestamp(time.time()))
    
    msg_str = (
        f"{timestamp} {msg[0]*100:.2f}% {task} done after "
        f"{msg[1]:.2f} minutes. Remaining {msg[2]:.2f} minutes."
    )
    
    if job_num < num_jobs:
        sys.stderr.write(msg_str + "\r")
    else:
        sys.stderr.write(msg_str + "\n")


"""
Implements Bagged Combinatorial Purged Cross-Validation (B-CPCV).
"""

import warnings
from collections import ChainMap
from typing import (
    Any, Dict, Optional, Union
)

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from sklearn.base import clone
from sklearn.ensemble import BaggingClassifier, BaggingRegressor
from sklearn.exceptions import ConvergenceWarning

from .combinatorial_purged import CombinatorialPurged

# For type hinting sklearn-like estimators
Estimator = Any

class BaggedCombinatorialPurged(CombinatorialPurged):
    """
    Bagged Combinatorial Purged Cross-Validation (B-CPCV).

    This class extends CPCV by applying a bagging (Bootstrap Aggregating)
    wrapper to the estimator for each combinatorial split. This helps
    to reduce variance and improve model stability.

    Parameters
    ----------
    n_splits : int
        Number of splits/groups to partition the data into.
    n_test_groups : int
        Size of the testing set in terms of groups.
    times : pd.Series or Dict[str, pd.Series]
        The timestamp series associated with the labels.
    embargo : float, default=0
        The embargo rate for purging.
    classifier : bool, default=True
        If True, use `sklearn.ensemble.BaggingClassifier`.
        If False, use `sklearn.ensemble.BaggingRegressor`.
    n_estimators : int, default=10
        The number of base estimators in the ensemble.
    max_samples : float, default=1.0
        The number of samples to draw to train each base estimator.
    max_features : float, default=1.0
        The number of features to draw to train each base estimator.
    bootstrap : bool, default=True
        Whether samples are drawn with replacement.
    bootstrap_features : bool, default=False
        Whether features are drawn with replacement.
    random_state : int, optional
        Controls the randomness of the bootstrapping.
    """

    def __init__(
        self,
        n_splits: int,
        n_test_groups: int,
        times: Union[pd.Series, Dict[str, pd.Series]],
        embargo: float = 0,
        classifier: bool = True,
        n_estimators: int = 10,
        max_samples: float = 1.0,
        max_features: float = 1.0,
        bootstrap: bool = True,
        bootstrap_features: bool = False,
        random_state: int = None
    ):
        """
        Initialize the BaggedCombinatorialPurged class.
        """
        super().__init__(n_splits, n_test_groups, times, embargo)
        self.classifier = classifier
        self.n_estimators = n_estimators
        self.max_samples = max_samples
        self.max_features = max_features
        self.bootstrap = bootstrap
        self.bootstrap_features = bootstrap_features
        self.random_state = random_state

    def _single_backtest_predictions(
        self,
        single_estimator: Estimator,
        single_times: pd.Series,
        single_data: pd.DataFrame,
        single_labels: pd.Series,
        single_weights: Optional[np.ndarray] = None,
        predict_probability: bool = False,
        n_jobs: int = 1
    ) -> Dict[str, np.ndarray]:
        """
        Obtain backtest predictions for all B-CPCV paths.

        This method overrides the parent by wrapping the `single_estimator`
        in a `BaggingClassifier` or `BaggingRegressor` before training.

        Parameters
        ----------
        single_estimator : Estimator
            The *base* estimator to be used inside the bagging ensemble.
        single_times : pd.Series
            The 'times' (info range) Series for this dataset.
        single_data : pd.DataFrame
            Data for the single dataset.
        single_labels : pd.Series
            Labels for the single dataset.
        single_weights : np.ndarray, optional
            Sample weights. Defaults to equal weights.
        predict_probability : bool, default=False
            If True, call `predict_proba()` instead of `predict()`.
            Only valid if `classifier=True`.
        n_jobs : int, default=1
            The number of jobs to run in parallel.

        Returns
        -------
        Dict[str, np.ndarray]
            A dictionary where keys are "Path {id}" and values are
            the contiguous arrays of out-of-sample predictions for each path.
        """
        self._validate_input(single_times, single_data)

        if single_weights is None:
            single_weights = np.ones(len(single_data))
            
        if predict_probability and not self.classifier:
            raise ValueError(
                "Cannot use predict_probability=True when classifier=False"
            )

        combinations_list, locations, split_segments = \
            self._combinations_and_path_locations_and_split_segments(single_data)

        def train_single_bagging_estimator(
            base_estimator_: Estimator,
            combinatorial_test_indices: np.ndarray
        ) -> Estimator:
            """Train one Bagging estimator for one C(n,k) split."""
            train_indices = self._get_train_indices(
                combinatorial_test_indices,
                single_times,
                continous_test_times=False
            )

            X_train = single_data.iloc[train_indices]
            y_train = single_labels.iloc[train_indices]
            weights_train = single_weights[train_indices]

            # --- Bagging Wrapper ---
            if self.classifier:
                bagging_estimator = BaggingClassifier(
                    estimator=base_estimator_,
                    n_estimators=self.n_estimators,
                    max_samples=self.max_samples,
                    max_features=self.max_features,
                    bootstrap=self.bootstrap,
                    bootstrap_features=self.bootstrap_features,
                    random_state=self.random_state,
                    n_jobs=n_jobs  # Parallelize bagging itself
                )
            else:
                bagging_estimator = BaggingRegressor(
                    estimator=base_estimator_,
                    n_estimators=self.n_estimators,
                    max_samples=self.max_samples,
                    max_features=self.max_features,
                    bootstrap=self.bootstrap,
                    bootstrap_features=self.bootstrap_features,
                    random_state=self.random_state,
                    n_jobs=n_jobs # Parallelize bagging itself
                )
            # --- End Bagging Wrapper ---

            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=ConvergenceWarning)
                try:
                    bagging_estimator.fit(X_train, y_train, sample_weight=weights_train)
                except (TypeError, ValueError):
                    # Fallback for models without sample_weight
                    bagging_estimator.fit(X_train, y_train)

            return bagging_estimator

        # 1. Train all C(n, k) *bagging* estimators in parallel
        # Note: We set n_jobs=1 for the *outer* parallel loop
        # and let the *inner* bagging estimator use `n_jobs`.
        # This avoids nested parallelization issues.
        
        # Determine parallelization strategy
        if n_jobs > 1 or n_jobs == -1:
            outer_n_jobs = n_jobs
            inner_n_jobs = 1 
            
            # Update bagging params to use inner_n_jobs
            if self.classifier:
                BaggingClassifier.__init__ = (
                    lambda self, **kwargs: 
                    super(BaggingClassifier, self).__init__(**kwargs, n_jobs=inner_n_jobs)
                )
            else:
                BaggingRegressor.__init__ = (
                    lambda self, **kwargs: 
                    super(BaggingRegressor, self).__init__(**kwargs, n_jobs=inner_n_jobs)
                )
        else:
            # Let bagging use all cores if outer loop is serial
            outer_n_jobs = 1
            inner_n_jobs = -1 # Use all
        
        combinatorial_trained_estimators = Parallel(n_jobs=outer_n_jobs)(
            delayed(train_single_bagging_estimator)(
                clone(single_estimator), test_indices
            )
            for test_indices in self._combinatorial_splits(
                combinations_list, split_segments
            )
        )
        
        # Restore default constructors
        BaggingClassifier.__init__ = BaggingClassifier.__init__
        BaggingRegressor.__init__ = BaggingRegressor.__init__


        # 2. Assemble predictions (this is fast, can be serial or parallel)
        def get_path_data(
            path_num: int,
            locs: List[Tuple[int, int]]
        ) -> Dict[str, np.ndarray]:
            """Assemble predictions for one path."""
            path_predictions = []
            for (group_idx, split_idx) in locs:
                test_indices_segment = split_segments[group_idx]
                X_test = single_data.iloc[test_indices_segment]
                estimator = combinatorial_trained_estimators[split_idx]

                if predict_probability:
                    preds = estimator.predict_proba(X_test)
                else:
                    preds = estimator.predict(X_test)

                path_predictions.append(preds)
            
            return {f"Path {path_num}": np.concatenate(path_predictions)}

        path_results = Parallel(n_jobs=outer_n_jobs)(
            delayed(get_path_data)(path_num, locs)
            for path_num, locs in locations.items()
        )
        
        paths_predictions = dict(ChainMap(*reversed(path_results)))
        return paths_predictions


"""
Utilities for creating publication-quality plots with Matplotlib
and Seaborn, using Times New Roman font and high DPI.
"""

import matplotlib.pyplot as plt
import seaborn as sns
from typing import Optional, List

def setup_publication_style() -> None:
    """
    Sets the global Matplotlib rcParams for a consistent,
    publication-quality (Times New Roman, 300 DPI) style.
    
    Call this function once at the beginning of your notebook.
    """
    
    # Check if Times New Roman is available
    try:
        plt.rc('font', family='Times New Roman')
    except:
        print("Warning: Times New Roman not found. Defaulting to serif.")
        plt.rc('font', family='serif')

    params = {
        'font.size': 12,
        'axes.labelsize': 12,
        'axes.titlesize': 14,
        'xtick.labelsize': 12,
        'ytick.labelsize': 12,
        'legend.fontsize': 12,
        'figure.dpi': 300,
        'savefig.dpi': 300,
        'savefig.transparent': True,
        'axes.grid': True,
        'grid.linestyle': '--',
        'grid.alpha': 0.5,
        'axes.edgecolor': 'black',
        'axes.linewidth': 1.2,
    }
    
    plt.rcParams.update(params)
    sns.set_style("whitegrid", params)
    print("Matplotlib style updated for publication.")


def apply_plot_style(
    ax: plt.Axes,
    title: str,
    xlabel: str,
    ylabel: str,
    legend_title: Optional[str] = None
) -> None:
    """
    Applies the standardized style to a specific Matplotlib Axes object.

    Parameters
    ----------
    ax : plt.Axes
        The Matplotlib axes to style.
    title : str
        The title for the plot.
    xlabel : str
        The label for the x-axis.
    ylabel : str
        The label for the y-axis.
    legend_title : str, optional
        Title for the legend, if any.
    """
    ax.set_title(title, fontsize=14, fontweight='bold')
    ax.set_xlabel(xlabel, fontsize=12)
    ax.set_ylabel(ylabel, fontsize=12)
    
    if ax.get_legend():
        ax.legend(title=legend_title, fontsize=12)


"""
RiskLabAI Backtest Validation Module

This module provides a suite of advanced cross-validation tools
for financial machine learning, specializing in methods that
prevent information leakage, such as Purged K-Fold and
Combinatorial Purged Cross-Validation (CPCV).
"""

from .cross_validator_interface import CrossValidator
from .kfold import KFold
from .walk_forward import WalkForward
from .purged_kfold import PurgedKFold
from .combinatorial_purged import CombinatorialPurged
from .adaptive_combinatorial_purged import AdaptiveCombinatorialPurged
from .bagged_combinatorial_purged import BaggedCombinatorialPurged
from .cross_validator_factory import CrossValidatorFactory
from .cross_validator_controller import CrossValidatorController

__all__ = [
    "CrossValidator",
    "KFold",
    "WalkForward",
    "PurgedKFold",
    "CombinatorialPurged",
    "AdaptiveCombinatorialPurged",
    "BaggedCombinatorialPurged",
    "CrossValidatorFactory",
    "CrossValidatorController"
]


"""
RiskLabAI Ensemble Module

Implements ensemble-related methods, such as the theoretical
accuracy of bagging classifiers.
"""

from .bagging_classifier_accuracy import bagging_classifier_accuracy

__all__ = [
    "bagging_classifier_accuracy"
]


"""
RiskLabAI PDE Solver Module

Implements a Deep BSDE (Backward Stochastic Differential Equation)
solver for various financial PDEs.
"""

from .equation import (
    Equation,
    PricingDefaultRisk,
    HJBLQ,
    BlackScholesBarenblatt,
    PricingDiffRate,
)
from .model import (
    TimeNet,
    Net1,
    MAB,
    SAB,
    ISAB,
    PMA,
    TimeNetForSet,
    DeepTimeSetTransformer,
    FBSNNNetwork,
    DeepBSDE,
    TimeDependentNetwork,
    TimeDependentNetworkMonteCarlo,
)
from .solver import (
    initialize_weights,
    FBSDESolver,
    FBSNNolver, # Note: Typo in original filename?
)

__all__ = [
    # Equations
    "Equation",
    "PricingDefaultRisk",
    "HJBLQ",
    "BlackScholesBarenblatt",
    "PricingDiffRate",
    
    # Models
    "TimeNet",
    "Net1",
    "MAB", "SAB", "ISAB", "PMA",
    "TimeNetForSet",
    "DeepTimeSetTransformer",
    "FBSNNNetwork",
    "DeepBSDE",
    "TimeDependentNetwork",
    "TimeDependentNetworkMonteCarlo",
    
    # Solvers
    "initialize_weights",
    "FBSDESolver",
    "FBSNNolver",
]


"""
Abstract Base Class for feature importance strategies.
"""

from abc import ABC, abstractmethod
import pandas as pd
from typing import Any

class FeatureImportanceStrategy(ABC):
    """
    Abstract Base Class for computing feature importance.

    Derived classes must implement the `compute` method.
    """

    @abstractmethod
    def compute(
        self, x: pd.DataFrame, y: pd.Series, **kwargs: Any
    ) -> pd.DataFrame:
        """
        Abstract method to compute feature importance.

        Parameters
        ----------
        x : pd.DataFrame
            The feature data.
        y : pd.Series
            The target data.
        **kwargs : Any
            Additional keyword arguments (e.g., sample_weights).

        Returns
        -------
        pd.DataFrame
            A DataFrame containing feature importances.
        """
        pass


"""
Defines the neural network architectures used by the PDE solvers.

Includes:
- Standard feed-forward networks (Net1, TimeNet)
- Set-Transformer components (MAB, SAB, ISAB, PMA)
- Deep BSDE models (DeepBSDE, FBSNNNetwork, TimeDependentNetwork)

Set-Transformer code based on: https://github.com/juho-lee/set_transformer
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from torch.nn import Module, Linear, BatchNorm1d, Tanh, ReLU
import math
from typing import List

# --- Set Transformer Components ---

class MAB(Module):
    """Multi-Head Attention Block (MAB)."""
    def __init__(self, dim_q: int, dim_k: int, dim_v: int, num_heads: int, ln: bool = False):
        super(MAB, self).__init__()
        self.dim_v = dim_v
        self.num_heads = num_heads
        self.fc_q = Linear(dim_q, dim_v)
        self.fc_k = Linear(dim_k, dim_v)
        self.fc_v = Linear(dim_k, dim_v)
        self.ln0 = nn.LayerNorm(dim_v) if ln else None
        self.ln1 = nn.LayerNorm(dim_v) if ln else None
        self.fc_o = Linear(dim_v, dim_v)

    def forward(self, q: Tensor, k: Tensor) -> Tensor:
        q_fc = self.fc_q(q)
        k_fc, v_fc = self.fc_k(k), self.fc_v(k)

        dim_split = self.dim_v // self.num_heads
        q_ = torch.cat(q_fc.split(dim_split, 2), 0)
        k_ = torch.cat(k_fc.split(dim_split, 2), 0)
        v_ = torch.cat(v_fc.split(dim_split, 2), 0)

        attention = torch.softmax(q_.bmm(k_.transpose(1, 2)) / math.sqrt(self.dim_v), 2)
        out = torch.cat((q_ + attention.bmm(v_)).split(q.size(0), 0), 2)
        
        out = self.ln0(out) if self.ln0 is not None else out
        out = out + F.relu(self.fc_o(out))
        out = self.ln1(out) if self.ln1 is not None else out
        return out

class SAB(Module):
    """Self-Attention Block (SAB)."""
    def __init__(self, dim_in: int, dim_out: int, num_heads: int, ln: bool = False):
        super(SAB, self).__init__()
        self.mab = MAB(dim_in, dim_in, dim_out, num_heads, ln=ln)

    def forward(self, x: Tensor) -> Tensor:
        return self.mab(x, x)

# ... (ISAB, PMA, TimeNetForSet, DeepTimeSetTransformer refactored similarly) ...

# --- Standard Networks ---

class TimeNet(Module):
    """Simple feed-forward network for time features."""
    def __init__(self, output_dim: int):
        super(TimeNet, self).__init__()
        self.layer1 = Linear(4, 100)
        self.layer2 = Linear(100, 150)
        self.layer3 = Linear(150, 200)
        self.layer4 = Linear(200, 300)
        self.layer5 = Linear(300, 200)
        self.layer6 = Linear(200, output_dim)
        self.tanh_stack = nn.Sequential(
            Tanh(), Tanh(), Tanh(), Tanh(), Tanh(), Tanh()
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.tanh_stack[0](self.layer1(x))
        x = self.tanh_stack[1](self.layer2(x))
        x = self.tanh_stack[2](self.layer3(x))
        x = self.tanh_stack[3](self.layer4(x))
        x = self.tanh_stack[4](self.layer5(x))
        x = self.tanh_stack[5](self.layer6(x))
        return x


class Net1(Module):
    """A simple Linear + BatchNorm layer."""
    def __init__(self, input_dim: int, output_dim: int):
        super(Net1, self).__init__()
        self.layer = Linear(input_dim, output_dim)
        self.bn = BatchNorm1d(output_dim)

    def forward(self, x: Tensor) -> Tensor:
        return self.layer(x)

# --- BSDE Solver Networks ---

class FBSNNNetwork(Module):
    """Feed-forward network for the FBSNN solver."""
    def __init__(self, layer_sizes: List[int]):
        super(FBSNNNetwork, self).__init__()
        self.n_layer = len(layer_sizes) - 1
        self.layers = nn.ModuleList([])
        
        for i in range(self.n_layer):
            self.layers.append(Linear(layer_sizes[i], layer_sizes[i + 1]))

    def forward(self, x: Tensor) -> Tensor:
        for i in range(self.n_layer):
            x = self.layers[i](x)
            if i < self.n_layer - 1:
                x = torch.sin(x)  # Use sin activation
        return x


class DeepBSDE(Module):
    """Network for the Deep BSDE method (one net per time step)."""
    def __init__(self, layer_sizes: List[int]):
        super(DeepBSDE, self).__init__()
        self.n_layer = len(layer_sizes) - 1
        self.layers = nn.ModuleList([])
        self.batch_layer = nn.ModuleList([])

        for i in range(self.n_layer):
            self.layers.append(Linear(layer_sizes[i], layer_sizes[i + 1], bias=False))
            self.batch_layer.append(BatchNorm1d(layer_sizes[i], eps=1e-06, momentum=0.01))
            
        self.batch_layer.append(BatchNorm1d(layer_sizes[-1], eps=1e-06, momentum=0.01))

    def forward(self, x: Tensor) -> Tensor:
        for i in range(self.n_layer - 1):
            x = self.layers[i](x)
            x = nn.ReLU()(x)
            
        x = self.layers[-1](x)
        return x

# ... (TimeDependentNetwork, TimeDependentNetworkMonteCarlo refactored similarly) ...

# import torch

# from torch.nn import Module, Linear, BatchNorm1d, Tanh
# class TimeNet(Module):
#     """
#     Neural network model for time dimension
#     """
#     def __init__(
#             self,
#             output_dim: int
#     ):
#         """
#         Initialize the neural network model with layers

#         :param output_dim: The output dimension of the neural network
#         :type output_dim: int
#         """
#         super(TimeNet, self).__init__()
#         self.layer1 = Linear(4, 100)
#         self.layer2 = Linear(100, 150)
#         self.layer3 = Linear(150, 200)
#         self.layer4 = Linear(200, 300)
#         self.layer5 = Linear(300, 200)
#         self.layer6 = Linear(200, output_dim)
#         self.tanh1 = Tanh()
#         self.tanh2 = Tanh()
#         self.tanh3 = Tanh()
#         self.tanh4 = Tanh()
#         self.tanh5 = Tanh()
#         self.tanh6 = Tanh()

#     def forward(
#             self,
#             x: torch.Tensor
#     ) -> torch.Tensor:
#         """
#         Forward propagation through the network.

#         :param x: Input tensor
#         :type x: torch.Tensor
#         :return: Output tensor
#         :rtype: torch.Tensor
#         """
#         x = self.layer1(x)
#         x = self.tanh1(x)
#         x = self.layer2(x)
#         x = self.tanh2(x)
#         x = self.layer3(x)
#         x = self.tanh3(x)
#         x = self.layer4(x)
#         x = self.tanh4(x)
#         x = self.layer5(x)
#         x = self.tanh5(x)
#         x = self.layer6(x)
#         x = self.tanh6(x)
#         return x


# class Net1(Module):
#     """
#     A class for defining a neural network with a single linear layer.
#     """
#     def __init__(
#             self,
#             input_dim: int,
#             output_dim: int
#     ):
#         """
#         Initialize the network with a single linear layer.

#         :param input_dim: Number of input features
#         :type input_dim: int
#         :param output_dim: Number of output features
#         :type output_dim: int
#         """
#         super(Net1, self).__init__()
#         self.layer = Linear(input_dim, output_dim)
#         self.bn = BatchNorm1d(output_dim)

#     def forward(
#             self,
#             x: torch.Tensor
#     ) -> torch.Tensor:
#         """
#         Forward propagation through the network.

#         :param x: Input tensor of shape (batch_size, input_dim)
#         :type x: torch.Tensor
#         :return: Output tensor of shape (batch_size, output_dim)
#         :rtype: torch.Tensor
#         """
#         x = self.layer(x)
#         return x
# import math
# import torch
# import torch.nn as nn
# import torch.nn.functional as F
# """
# For implementing Deep-Time-SetTransformer we have used code written by authors of Set-Transformer available at : https://github.com/juho-lee/set_transformer
# """

# class MAB(nn.Module):
#     def __init__(
#         self,
#         dim_q: int,
#         dim_k: int,
#         dim_v: int,
#         num_heads: int,
#         ln: bool = False
#     ):
#         """
#         Multi-Head Self Attention Block.

#         :param dim_q: Dimension of query
#         :param dim_k: Dimension of key
#         :param dim_v: Dimension of value
#         :param num_heads: Number of attention heads
#         :param ln: Whether to use Layer Normalization
#         """
#         super(MAB, self).__init__()
#         self.dim_v = dim_v
#         self.num_heads = num_heads
#         self.fc_q = nn.Linear(dim_q, dim_v)
#         self.fc_k = nn.Linear(dim_k, dim_v)
#         self.fc_v = nn.Linear(dim_k, dim_v)
#         if ln:
#             self.ln0 = nn.LayerNorm(dim_v)
#             self.ln1 = nn.LayerNorm(dim_v)
#         self.fc_o = nn.Linear(dim_v, dim_v)

#     def forward(
#         self,
#         q: torch.Tensor,
#         k: torch.Tensor
#     ) -> torch.Tensor:
#         """
#         Forward propagation.

#         :param q: Query tensor
#         :param k: Key tensor
#         :return: Output tensor
#         """
#         q = self.fc_q(q)
#         k, v = self.fc_k(k), self.fc_v(k)

#         dim_split = self.dim_v // self.num_heads
#         q_ = torch.cat(q.split(dim_split, 2), 0)
#         k_ = torch.cat(k.split(dim_split, 2), 0)
#         v_ = torch.cat(v.split(dim_split, 2), 0)

#         a = torch.softmax(q_.bmm(k_.transpose(1,2))/math.sqrt(self.dim_v), 2)
#         o = torch.cat((q_ + a.bmm(v_)).split(q.size(0), 0), 2)
#         o = o if getattr(self, 'ln0', None) is None else self.ln0(o)
#         o = o + F.relu(self.fc_o(o))
#         o = o if getattr(self, 'ln1', None) is None else self.ln1(o)
#         return o

# class SAB(nn.Module):
#     def __init__(
#         self,
#         dim_in: int,
#         dim_out: int,
#         num_heads: int,
#         ln: bool = False
#     ):
#         """
#         Self Attention Block.

#         :param dim_in: Input dimension
#         :param dim_out: Output dimension
#         :param num_heads: Number of attention heads
#         :param ln: Whether to use Layer Normalization
#         """
#         super(SAB, self).__init__()
#         self.mab = MAB(dim_in, dim_in, dim_out, num_heads, ln=ln)

#     def forward(
#         self,
#         x: torch.Tensor
#     ) -> torch.Tensor:
#         """
#         Forward propagation.

#         :param x: Input tensor
#         :return: Output tensor
#         """
#         return self.mab(x, x)

# class ISAB(nn.Module):
#     def __init__(
#         self,
#         dim_in: int,
#         dim_out: int,
#         num_heads: int,
#         num_inds: int,
#         ln: bool = False
#     ):
#         """
#         Induced Self Attention Block.

#         :param dim_in: Input dimension
#         :param dim_out: Output dimension
#         :param num_heads: Number of attention heads
#         :param num_inds: Number of inducing points
#         :param ln: Whether to use Layer Normalization
#         """
#         super(ISAB, self).__init__()
#         self.i = nn.Parameter(torch.Tensor(1, num_inds, dim_out))
#         nn.init.xavier_uniform_(self.i)
#         self.mab0 = MAB(dim_out, dim_in, dim_out, num_heads, ln=ln)
#         self.mab1 = MAB(dim_in, dim_out, dim_out, num_heads, ln=ln)

#     def forward(
#         self,
#         x: torch.Tensor
#     ) -> torch.Tensor:
#         """
#         Forward propagation.

#         :param x: Input tensor
#         :return: Output tensor
#         """
#         h = self.mab0(self.i.repeat(x.size(0), 1, 1), x)
#         return self.mab1(x, h)

# class PMA(nn.Module):
#     def __init__(
#         self,
#         dim: int,
#         num_heads: int,
#         num_seeds: int,
#         ln: bool = False
#     ):
#         """
#         Pooling Multihead Attention.

#         :param dim: Dimension of input and output
#         :param num_heads: Number of attention heads
#         :param num_seeds: Number of seed vectors
#         :param ln: Whether to use Layer Normalization
#         """
#         super(PMA, self).__init__()
#         self.s = nn.Parameter(torch.Tensor(1, num_seeds, dim))
#         nn.init.xavier_uniform_(self.s)
#         self.mab = MAB(dim, dim, dim, num_heads, ln=ln)

#     def forward(
#         self,
#         x: torch.Tensor
#     ) -> torch.Tensor:
#         """
#         Forward propagation.

#         :param x: Input tensor
#         :return: Output tensor
#         """
#         return self.mab(self.s.repeat(x.size(0), 1, 1), x)
# import torch
# import torch.nn as nn


# class TimeNetForSet(nn.Module):
#     """
#     Neural network model for time dimension.

#     Args:
#         in_features (int): The input features dimension. Default is 1.
#         out_features (int): The output features dimension. Default is 64.
#     """
#     def __init__(self,
#                  in_features: int = 1,
#                  out_features: int = 64):
#         super(TimeNetForSet, self).__init__()
#         self.feature = nn.Linear(in_features=in_features, out_features=out_features)
#         self.layer1 = nn.Linear(1, 10)
#         self.layer2 = nn.Linear(10, 10)
#         self.layer3 = nn.Linear(10, 10)
#         self.layer4 = nn.Linear(10, out_features)
#         self.relu1 = nn.ReLU()
#         self.relu2 = nn.ReLU()
#         self.relu3 = nn.ReLU()

#     def forward(self,
#                 t: torch.Tensor,
#                 x: torch.Tensor) -> torch.Tensor:
#         """
#         Forward pass of the network.

#         Args:
#             t (torch.Tensor): Input tensor for time dimension.
#             x (torch.Tensor): Input tensor for features.

#         Returns:
#             torch.Tensor: Output tensor.
#         """
#         t = self.relu1(self.layer1(t))
#         t = self.relu2(self.layer2(t))
#         t = self.relu3(self.layer3(t))
#         t = self.layer4(t)
#         x = self.feature(x)

#         x = (torch.exp(t)) * x

#         return x

#     def freeze(self):
#         """Freezes the feature parameters."""
#         for param in self.feature.parameters():
#             param.requires_grad = False


# class DeepTimeSetTransformer(nn.Module):
#     def __init__(self, input_dim: int):
#         super(DeepTimeSetTransformer, self).__init__()
#         self.feature_extractor = nn.Sequential(
#             nn.Linear(in_features=1, out_features=64),
#             nn.ReLU(),
#             nn.Linear(in_features=64, out_features=64),
#             nn.ReLU(),
#             nn.Linear(in_features=64, out_features=64),
#             nn.ReLU(),
#             nn.Linear(in_features=64, out_features=64),
#         )
#         self.time_layer1 = TimeNetForSet(in_features=input_dim, out_features=32)
#         self.time_layer2 = TimeNetForSet(in_features=32, out_features=32)
#         self.time_layer3 = TimeNetForSet(in_features=32, out_features=32)
#         self.time_layer4 = TimeNetForSet(in_features=32, out_features=32)
#         self.time_layer5 = TimeNetForSet(in_features=32, out_features=1)

#         self.layer1 = nn.Linear(in_features=input_dim, out_features=32)
#         self.layer2 = nn.Linear(in_features=32, out_features=32)
#         self.layer3 = nn.Linear(in_features=32, out_features=32)
#         self.layer4 = nn.Linear(in_features=32, out_features=32)
#         self.layer5 = nn.Linear(in_features=32, out_features=32)

#         self.regressor = nn.Sequential(
#             PMA(dim=32, num_heads=4, num_seeds=1),
#         )

#     def forward(self,
#                 t: torch.Tensor,
#                 x: torch.Tensor) -> torch.Tensor:
#         """
#         Forward pass of the network.

#         Args:
#             t (torch.Tensor): Input tensor for time dimension.
#             x (torch.Tensor): Input tensor for features.

#         Returns:
#             torch.Tensor: Output tensor.
#         """
#         x = self.layer1(x)
#         x = nn.ReLU()(x)
#         x = x - torch.mean(x, 2, keepdim=True)
#         x = self.layer2(x)

#         x = nn.ReLU()(x)
#         x = x - torch.mean(x, 2, keepdim=True)
#         x = self.layer3(x)

#         x = nn.ReLU()(x)
#         x = x - torch.mean(x, 2, keepdim=True)
#         x = self.layer4(x)
#         x = nn.ReLU()(x)
#         x = x - torch.mean(x, 2, keepdim=True)
#         x = self.layer5(x)
#         output = nn.ReLU()(x)

#         output = self.regressor(output)
#         output = torch.squeeze(output)

#         output = self.time_layer2(t, output)
#         output = nn.ReLU()(output)
#         output = self.time_layer3(t, output)
#         output = nn.ReLU()(output)
#         output = self.time_layer4(t, output)
#         output = nn.ReLU()(output)
#         output = self.time_layer5(t, output)


#         return output
# import torch
# import torch.nn as nn



# class FBSNNNetwork(nn.Module):
#     def __init__(
#             self,
#             layersize: list[int]
#             ):
#         """
#         Initializes a neural network with multiple blocks.

#         Args:
#         - indim (int): input dimension
#         - layersize (List[int]): list of sizes of hidden layers
#         - outdim (int): output dimension
#         """
#         super(FBSNNNetwork, self).__init__()

#         # initialize first set of layers

#         self.n_layer = len(layersize) - 1
#         # create input layer

#         self.layers = nn.ModuleList([])
#         # create hidden layers
#         for i in range(len(layersize) - 1):
#           self.layers.append(nn.Linear(layersize[i], layersize[i + 1]))

#     def forward(
#             self,
#             x: torch.Tensor) -> torch.Tensor:
#         """
#         Passes the input through the neural network.

#         Args:
#         - x (torch.Tensor): input tensor

#         Returns:
#         - torch.Tensor: output tensor
#         """

#         for i in range(self.n_layer):



#             # apply layer to tensor
#             x = self.layers[i](x)
#             if i < self.n_layer - 1 :
#               x =torch.sin(x)

#         # apply output layer to tensor


#         return x


# class DeepBSDE(nn.Module):
#     def __init__(
#             self,
#             layersize: list[int]
#             ):
#         """
#         Initializes a neural network with multiple blocks.

#         Args:
#         - indim (int): input dimension
#         - layersize (List[int]): list of sizes of hidden layers
#         - outdim (int): output dimension
#         """
#         super(DeepBSDE, self).__init__()

#         # initialize first set of layers

#         self.n_layer = len(layersize) - 1
#         # create input layer

#         self.layers = nn.ModuleList([])
#         self.batch_layer = nn.ModuleList([])
#         # create hidden layers
#         for i in range(len(layersize) - 1):
#           self.layers.append(nn.Linear(layersize[i], layersize[i + 1] , bias=False))
#           self.batch_layer.append(nn.BatchNorm1d(layersize[i] , eps=1e-06, momentum=0.01))


#         self.batch_layer.append(nn.BatchNorm1d(layersize[-1] , eps=1e-06, momentum=0.01))

#     def forward(
#             self,
#             x: torch.Tensor) -> torch.Tensor:
#         """
#         Passes the input through the neural network.

#         Args:
#         - x (torch.Tensor): input tensor

#         Returns:
#         - torch.Tensor: output tensor
#         """
#         #x = self.batch_layer[0](x)
#         for i in range(self.n_layer - 1):

#             # apply layer to tensor
#             x = self.layers[i](x)
#             #x = self.batch_layer[i+1](x)


#             x =nn.ReLU()(x)


#         # apply output layer to tensor
#         x = self.layers[-1](x)
#         #x = self.batch_layer[-1](x)


#         return x


# class TimeDependentNetwork(nn.Module):
#     def __init__(
#             self,
#             indim: int,
#             layersize: list[int],
#             outdim: int):
#         """
#         Initializes a neural network with multiple blocks.

#         Args:
#         - indim (int): input dimension
#         - layersize (List[int]): list of sizes of hidden layers
#         - outdim (int): output dimension
#         """
#         super(TimeDependentNetwork, self).__init__()

#         # initialize first set of layers
#         self.n_layer = len(layersize)
#         self.layers = nn.ModuleList([])
#         self.time_layer = nn.ModuleList([])
#         self.batch_layer = nn.ModuleList([])

#         # create input layer
#         self.layers.append(Net1(indim, layersize[0]))
#         self.time_layer.append(TimeNet(indim))
#         self.batch_layer.append(nn.BatchNorm1d(indim , eps=1e-06, momentum=0.01))
#         self.batch_layer.append(nn.BatchNorm1d(layersize[0],eps=1e-06, momentum=0.01))

#         # create hidden layers
#         for i in range(len(layersize) - 1):
#             self.layers.append(Net1(layersize[i], layersize[i + 1]))
#             self.time_layer.append(TimeNet(layersize[i]))
#             self.batch_layer.append(nn.BatchNorm1d(layersize[i + 1] , eps=1e-06, momentum=0.01))

#         # create output layer
#         self.time_layer.append(TimeNet(outdim))
#         self.linear = nn.Linear(layersize[-1], outdim)
#         self.batch_layer.append(nn.BatchNorm1d(outdim , eps=1e-06, momentum=0.01))


#     def forward(
#             self,
#             t: torch.Tensor,
#             x: torch.Tensor) -> torch.Tensor:
#         """
#         Passes the input through the neural network.

#         Args:
#         - t (torch.Tensor): tensor containing time information
#         - x (torch.Tensor): input tensor

#         Returns:
#         - torch.Tensor: output tensor
#         """
#         for i in range(self.n_layer):
#             # apply time-based weights to input tensor
#             time = self.time_layer[i](torch.cat((t, t**2, t**3, torch.exp(t)), 1))
#             x = self.batch_layer[i](x)


#             x = x * (1 + time)

#             # apply layer to tensor
#             x = self.layers[i](x)

#             x = torch.sin(x)

#         # apply output layer to tensor

#         x = self.linear(x)
#         #x = self.batch_layer[-1](x)

#         return x

# class TimeDependentNetworkMonteCarlo(nn.Module):
#     def __init__(
#             self,
#             indim: int,
#             layersize: list[int],
#             outdim: int,
#             sigma: float):
#         """
#         Initializes a neural network with multiple blocks.

#         Args:
#         - indim (int): input dimension
#         - layersize (List[int]): list of sizes of hidden layers
#         - outdim (int): output dimension
#         - sigma (float) : volatility
#         """
#         super(TimeDependentNetworkMonteCarlo, self).__init__()

#         # initialize first set of layers
#         self.n_layer = len(layersize)
#         self.layers = nn.ModuleList([])
#         self.time_layer = nn.ModuleList([])
#         self.batch_layer = nn.ModuleList([])

#         # create input layer
#         self.layers.append(Net1(indim, layersize[0]))
#         self.time_layer.append(TimeNet(indim))
#         self.batch_layer.append(nn.BatchNorm1d(indim))
#         self.batch_layer.append(nn.BatchNorm1d(layersize[0]))
#         self.sigma = sigma

#         # create hidden layers
#         for i in range(len(layersize) - 1):
#             self.layers.append(Net1(layersize[i], layersize[i + 1]))
#             self.time_layer.append(TimeNet(layersize[i]))
#             self.batch_layer.append(nn.BatchNorm1d(layersize[i + 1]))

#         # create output layer
#         self.time_layer.append(TimeNet(outdim))
#         self.linear = nn.Linear(layersize[-1], outdim)

#     def forward(
#             self,
#             t: torch.Tensor,
#             x: torch.Tensor,
#             y) -> torch.Tensor:
#         """
#         Passes the input through the neural network.

#         Args:
#         - t (torch.Tensor): tensor containing time information
#         - x (torch.Tensor): input tensor

#         Returns:
#         - torch.Tensor: output tensor
#         """
#         x_prim = x

#         for i in range(self.n_layer):
#             # apply time-based weights to input tensor
#             time = self.time_layer[i](torch.cat((t, t**2, t**3, torch.exp(t)), 1))

#             x = x * (1 + time)

#             # apply layer to tensor
#             x = self.layers[i](x)
#             x = nn.ReLU()(x)

#         # apply output layer to tensor
#         time = self.time_layer[self.n_layer](torch.cat((t, t**2, t**3, torch.exp(t)), 1))
#         x = (self.sigma * x_prim * self.linear(x) + (1 - time) * y)

#         return x


"""
Computes Clustered Mean Decrease Impurity (MDI) feature importance.
"""

import pandas as pd
import numpy as np
from sklearn.ensemble import BaseEnsemble
from typing import Dict, List, Any
from .feature_importance_strategy import FeatureImportanceStrategy

class ClusteredFeatureImportanceMDI(FeatureImportanceStrategy):
    """
    Computes Clustered MDI feature importance.

    Aggregates MDI importance for pre-defined clusters of features.
    """

    def __init__(
        self,
        classifier: BaseEnsemble,
        clusters: Dict[str, List[str]],
    ):
        """
        Initialize the strategy.

        Parameters
        ----------
        classifier : BaseEnsemble
            An *untrained* scikit-learn ensemble model.
        clusters : Dict[str, List[str]]
            Dictionary mapping cluster names to lists of feature names.
        """
        if not hasattr(classifier, 'estimators_'):
            raise TypeError("Classifier must be an ensemble (e.g., RandomForest).")
        self.classifier = classifier
        self.clusters = clusters

    def _group_mean_std(
        self, dataframe: pd.DataFrame, clusters: Dict[str, List[str]]
    ) -> pd.DataFrame:
        """
        Calculate the mean and standard deviation for cluster importances.
        """
        output = pd.DataFrame(columns=["Mean", "StandardDeviation"])

        for cluster_name, feature_names in clusters.items():
            # Sum importance for all features in the cluster
            cluster_data = dataframe[feature_names].sum(axis=1)
            
            cluster_mean = cluster_data.mean()
            cluster_std = cluster_data.std()
            
            output.loc[f"C_{cluster_name}", "Mean"] = cluster_mean
            output.loc[f"C_{cluster_name}", "StandardDeviation"] = (
                cluster_std * (cluster_data.shape[0] ** -0.5)
            )
        return output

    def compute(self, x: pd.DataFrame, y: pd.Series, **kwargs: Any) -> pd.DataFrame:
        """
        Compute Clustered MDI feature importance.

        Parameters
        ----------
        x : pd.DataFrame
            The feature data.
        y : pd.Series
            The target data.
        **kwargs : Any
            Keyword arguments for the classifier's `fit` method
            (e.g., `sample_weight`).

        Returns
        -------
        pd.DataFrame
            DataFrame with "Mean" and "StandardDeviation" of importance
            for each *cluster*.
        """
        train_sample_weights = kwargs.get('sample_weight')
        
        # Fit the classifier
        self.classifier.fit(x, y, sample_weight=train_sample_weights)
        
        # Get importance from each tree
        importances_dict = {
            i: tree.feature_importances_
            for i, tree in enumerate(self.classifier.estimators_)
        }
        importances_df = pd.DataFrame.from_dict(importances_dict, orient="index")
        
        if hasattr(self.classifier, 'feature_names_in_'):
             importances_df.columns = self.classifier.feature_names_in_
        else:
             importances_df.columns = x.columns

        # Replace 0 with NaN
        importances_df.replace(0, np.nan, inplace=True)

        # Group by cluster
        aggregated_importances = self._group_mean_std(importances_df, self.clusters)
        
        # Normalize
        aggregated_importances /= aggregated_importances["Mean"].sum()
        return aggregated_importances


"""
Implements the Probability of Backtest Overfitting (PBO) calculation.
"""

from typing import Tuple, Callable, List, Optional
from itertools import combinations
import numpy as np
from numba import jit
from joblib import Parallel, delayed

@jit(nopython=True)
def sharpe_ratio(
    returns: np.ndarray, risk_free_rate: float = 0.0
) -> float:
    """
    Calculate the Sharpe Ratio (Numba-optimized).

    Parameters
    ----------
    returns : np.ndarray
        An array of returns.
    risk_free_rate : float, default=0.0
        The risk-free rate.

    Returns
    -------
    float
        The calculated Sharpe Ratio.
    """
    excess_returns = returns - risk_free_rate
    std_dev = np.std(excess_returns)
    
    if std_dev == 0.0:
        return 0.0
        
    return np.mean(excess_returns) / std_dev

def performance_evaluation(
    train_partition: np.ndarray,
    test_partition: np.ndarray,
    n_strategies: int,
    metric: Callable[[np.ndarray, float], float],
    risk_free_return: float,
) -> Tuple[bool, float]:
    """
    Evaluate strategy performance on train/test splits.

    This function finds the best strategy on the training partition and
    then ranks its performance on the test partition.

    Parameters
    ----------
    train_partition : np.ndarray
        Training data partition (T_train x N_strategies).
    test_partition : np.ndarray
        Testing data partition (T_test x N_strategies).
    n_strategies : int
        Number of strategies (N).
    metric : Callable
        Metric function (e.g., `sharpe_ratio`).
    risk_free_return : float
        Risk-free return for the metric.

    Returns
    -------
    Tuple[bool, float]
        - is_overfit: True if the best in-sample strategy is not
                      in the top half of out-of-sample strategies.
        - logit_value: The logit-transformed relative rank.
    """
    # 1. Find best strategy on training data
    evaluate_train = [
        metric(train_partition[:, i], risk_free_return)
        for i in range(n_strategies)
    ]
    best_strategy_idx = np.argmax(evaluate_train)

    # 2. Evaluate all strategies on test data
    evaluate_test = [
        metric(test_partition[:, i], risk_free_return)
        for i in range(n_strategies)
    ]

    # 3. Find rank of the best_strategy in the test set
    # Sort test evaluations and find where best_strategy_idx lands
    ranks = np.argsort(np.argsort(evaluate_test))
    rank_of_best_is_strategy = ranks[best_strategy_idx] + 1  # 1-based rank

    # 4. Calculate relative rank (omega_bar) and logit
    w_bar = rank_of_best_is_strategy / (n_strategies + 1)
    logit_value = np.log(w_bar / (1 - w_bar))

    # Overfit if logit is <= 0 (i.e., rank is in bottom half)
    return logit_value <= 0.0, logit_value


def probability_of_backtest_overfitting(
    performances: np.ndarray,
    n_partitions: int = 16,
    risk_free_return: float = 0.0,
    metric: Optional[Callable[[np.ndarray, float], float]] = None,
    n_jobs: int = 1,
) -> Tuple[float, np.ndarray]:
    r"""
    Compute the Probability of Backtest Overfitting (PBO).

    PBO is the frequency with which the best in-sample (IS) strategy
    underperforms relative to the median out-of-sample (OOS) performance.

    It uses combinatorial splits of the data into train/test partitions.
    Total combinations = C(S, S/2), where S = n_partitions.

    .. math::
        \left(\begin{array}{c}
        S \\
        S / 2
        \end{array}\right) = \prod_{i=0}^{S / 2^{-1}} \frac{S-i}{S / 2-i}

    Parameters
    ----------
    performances : np.ndarray
        Matrix of T observations x N strategies.
    n_partitions : int, default=16
        Number of partitions (S). Must be even.
    risk_free_return : float, default=0.0
        Risk-free return for calculating the metric (e.g., Sharpe ratio).
    metric : Callable, optional
        Metric function (e.g., `sharpe_ratio`). Defaults to the
        Numba-jitted `sharpe_ratio` in this module.
    n_jobs : int, default=1
        Number of parallel jobs for `joblib`.

    Returns
    -------
    Tuple[float, np.ndarray]
        - pbo: The Probability of Backtest Overfitting (0 to 1).
        - logit_values: Array of logit values from all combinations.
    """
    if n_partitions % 2 != 0:
        raise ValueError("Number of partitions must be even.")

    if metric is None:
        metric = sharpe_ratio

    _, n_strategies = performances.shape
    partitions = np.array_split(performances, n_partitions)
    partition_indices = list(range(n_partitions))
    
    # Get all combinations of training partition indices
    partition_combinations_indices = list(
        combinations(partition_indices, n_partitions // 2)
    )

    results = Parallel(n_jobs=n_jobs)(
        delayed(performance_evaluation)(
            np.concatenate([partitions[i] for i in train_indices], axis=0),
            np.concatenate(
                [
                    partitions[i]
                    for i in partition_indices
                    if i not in train_indices
                ],
                axis=0,
            ),
            n_strategies,
            metric,
            risk_free_return,
        )
        for train_indices in partition_combinations_indices
    )

    results_arr = np.array(results)

    pbo = results_arr[:, 0].mean()
    logit_values = results_arr[:, 1]

    return pbo, logit_values


"""
Implements fractional differentiation of time series, including both
expanding-window (standard) and fixed-width-window (FFD) methods.

Includes functions to find the minimum differentiation factor 'd' that
results in a stationary series.

Reference:
    De Prado, M. (2018) Advances in financial machine learning.
    John Wiley & Sons, Chapter 5.
"""

import numpy as np
import pandas as pd
from statsmodels.tsa.stattools import adfuller
from typing import Tuple, Optional

def calculate_weights_std(degree: float, size: int) -> np.ndarray:
    """
    Compute weights for standard (expanding window) fractional differentiation.

    Reference:
        Snippet 5.2, Page 82.

    Parameters
    ----------
    degree : float
        The degree of differentiation (d).
    size : int
        The number of weights to generate (e.g., the series length).

    Returns
    -------
    np.ndarray
        An array of weights, ordered for dot product (w_0 at the end).
    """
    weights = [1.0]
    for k in range(1, size):
        weight = -weights[-1] / k * (degree - k + 1)
        weights.append(weight)
    
    # Reverse for dot product: [w_k, ..., w_1, w_0]
    return np.array(weights[::-1]).reshape(-1, 1)

def calculate_weights_ffd(degree: float, threshold: float = 1e-5) -> np.ndarray:
    """
    Compute weights for Fixed-Width Window Fractional Differentiation (FFD).

    The weights are generated until they fall below the `threshold`.

    Reference:
        Snippet 5.3, Page 83.

    Parameters
    ----------
    degree : float
        The degree of differentiation (d).
    threshold : float, default=1e-5
        The minimum absolute weight to include.

    Returns
    -------
    np.ndarray
        An array of weights, ordered for dot product (w_0 at the end).
    """
    weights = [1.0]
    k = 1
    while abs(weights[-1]) >= threshold:
        weight = -weights[-1] / k * (degree - k + 1)
        weights.append(weight)
        k += 1
    
    # Reverse for dot product and drop w_0 (which is 1)
    return np.array(weights[::-1]).reshape(-1, 1)


def fractional_difference_std(
    series: pd.DataFrame,
    degree: float,
    threshold: float = 0.01
) -> pd.DataFrame:
    """
    Compute the standard (expanding window) fractionally differentiated series.

    This method uses all available past data for each calculation.

    Reference:
        Snippet 5.2, Page 82.

    Parameters
    ----------
    series : pd.DataFrame
        DataFrame of time series (one column per series).
    degree : float
        The degree of differentiation (d).
    threshold : float, default=0.01
        The weight-loss threshold to determine the minimum number
        of observations required (warm-up period).

    Returns
    -------
    pd.DataFrame
        DataFrame of fractionally differentiated series.
    """
    # 1. Compute weights for the full series
    weights = calculate_weights_std(degree, series.shape[0])
    
    # 2. Determine warm-up period
    weights_cumsum_abs = np.cumsum(np.abs(weights))
    weights_cumsum_abs /= weights_cumsum_abs[-1]
    skip = np.searchsorted(weights_cumsum_abs, threshold)
    
    result_df = pd.DataFrame(index=series.index, columns=series.columns)

    for name in series.columns:
        series_ffill = series[[name]].fillna(method="ffill").dropna()
        if series_ffill.empty:
            continue
            
        series_np = series_ffill.to_numpy() # More efficient

        for iloc in range(skip, series_np.shape[0]):
            # Get the relevant window of data and weights
            window_data = series_np[:iloc + 1]
            window_weights = weights[-(iloc + 1):]
            
            # Dot product
            result_df.loc[series_ffill.index[iloc], name] = np.dot(
                window_weights.T, window_data
            )[0, 0]

    return result_df.dropna()


def fractional_difference_fixed(
    series: pd.DataFrame,
    degree: float,
    threshold: float = 1e-5
) -> pd.DataFrame:
    """
    Compute the Fixed-Width Window (FFD) fractionally differentiated series.

    This method uses a fixed-width window, which is more efficient and
    less prone to memory biases.

    Reference:
        Snippet 5.3, Page 83.

    Parameters
    ----------
    series : pd.DataFrame
        DataFrame of time series (one column per series).
    degree : float
        The degree of differentiation (d).
    threshold : float, default=1e-5
        Threshold to determine the fixed window width.

    Returns
    -------
    pd.DataFrame
        DataFrame of fractionally differentiated series.
    """
    weights = calculate_weights_ffd(degree, threshold)
    width = len(weights)
    
    result_df = pd.DataFrame(index=series.index, columns=series.columns)

    for name in series.columns:
        series_ffill = series[[name]].fillna(method="ffill").dropna()
        if series_ffill.empty:
            continue
            
        series_np = series_ffill.to_numpy()

        for iloc in range(width, series_np.shape[0]):
            window_data = series_np[iloc - width : iloc + 1]
            
            result_df.loc[series_ffill.index[iloc], name] = np.dot(
                weights.T, window_data
            )[0, 0]
            
    return result_df.dropna()

def fractional_difference_fixed_single(
    series: pd.Series,
    degree: float,
    threshold: float = 1e-5
) -> pd.Series:
    """
    Compute the FFD series for a single `pd.Series`.

    Parameters
    ----------
    series : pd.Series
        Time series.
    degree : float
        The degree of differentiation (d).
    threshold : float, default=1e-5
        Threshold to determine the fixed window width.

    Returns
    -------
    pd.Series
        The fractionally differentiated series.
    """
    weights = calculate_weights_ffd(degree, threshold)
    width = len(weights)
    
    series_ffill = series.fillna(method="ffill").dropna()
    series_np = series_ffill.to_numpy()
    
    result_series = pd.Series(index=series.index, dtype="float64")

    for iloc in range(width, series_np.shape[0]):
        window_data = series_np[iloc - width : iloc + 1]
        
        result_series.loc[series_ffill.index[iloc]] = np.dot(
            weights.T, window_data
        )[0]

    return result_series.dropna()


def plot_weights(
    degree_range: Tuple[float, float],
    number_degrees: int,
    size: int
) -> None:
    """
    Plot the weights of fractionally differentiated series for various degrees.

    Parameters
    ----------
    degree_range : Tuple[float, float]
        (min_degree, max_degree) to plot.
    number_degrees : int
        Number of 'd' values to plot in the range.
    size : int
        Length of the time series (number of weights).
    """
    weights_df = pd.DataFrame()
    for degree in np.linspace(degree_range[0], degree_range[1], number_degrees):
        degree = round(degree, 2)
        weights = calculate_weights_std(degree, size)
        weights_df[degree] = pd.Series(weights.flatten(), index=range(size - 1, -1, -1))

    ax = weights_df.plot()
    ax.set_xlabel("Lag")
    ax.set_ylabel("Weight")
    ax.set_title("Fractional Differentiation Weights")
    ax.legend(title="Degree (d)")
    ax.get_figure().show()


def find_optimal_ffd_simple(
    input_series: pd.DataFrame,
    p_value_threshold: float = 0.05
) -> pd.DataFrame:
    """
    Find the minimum 'd' that passes the ADF test, for a range of d.

    This function tests `d` in 11 steps from 0 to 1.

    Reference:
        Snippet 5.4, Page 85.

    Parameters
    ----------
    input_series : pd.DataFrame
        DataFrame containing a 'close' column.
    p_value_threshold : float, default=0.05
        The ADF test p-value to pass.

    Returns
    -------
    pd.DataFrame
        DataFrame of ADF test results for each 'd'.
    """
    results_list = []
    
    # Resample to daily to ensure consistent lags
    series_daily = np.log(input_series[['close']]).resample('1D').last()
    
    for d in np.linspace(0, 1, 11):
        differentiated = fractional_difference_fixed(
            series_daily, d, threshold=0.01
        ).dropna()
        
        if differentiated.empty:
            continue
            
        corr = np.corrcoef(
            series_daily.loc[differentiated.index, 'close'],
            differentiated['close']
        )[0, 1]
        
        adf_result = adfuller(
            differentiated['close'], maxlag=1, regression='c', autolag=None
        )
        
        results_list.append(
            {
                'd': d,
                'adfStat': adf_result[0],
                'pVal': adf_result[1],
                'lags': adf_result[2],
                'nObs': adf_result[3],
                '95% conf': adf_result[4]['5%'],
                'corr': corr
            }
        )
    
    return pd.DataFrame(results_list).set_index('d')


def fractionally_differentiated_log_price(
    input_series: pd.Series,
    threshold: float = 1e-5,
    step: float = 0.01,
    p_value_threshold: float = 0.05
) -> pd.Series:
    """
    Find the minimum 'd' that makes a log-price series stationary.

    This function iteratively increases 'd' by 'step' until the ADF test
    p-value falls below `p_value_threshold`.

    Parameters
    ----------
    input_series : pd.Series
        Time series of prices.
    threshold : float, default=1e-5
        The weight threshold for FFD.
    step : float, default=0.01
        The increment for testing 'd'.
    p_value_threshold : float, default=0.05
        The significance level for the ADF test.

    Returns
    -------
    pd.Series
        The fractionally differentiated series with the minimum
        stationary-passing 'd'.
    """
    log_price = np.log(input_series)
    degree = 0.0
    p_value = 1.0
    
    differentiated_series = None

    while p_value > p_value_threshold:
        degree += step
        if degree > 2.0: # Safety break
             raise ValueError("Failed to find stationary 'd' < 2.0")
             
        differentiated = fractional_difference_fixed_single(
            log_price, degree, threshold=threshold
        ).dropna()
        
        if differentiated.empty:
            continue # Not enough data for this 'd'
        
        adf_test = adfuller(
            differentiated, maxlag=1, regression='c', autolag=None
        )
        p_value = adf_test[1]
        
        if differentiated_series is None:
            differentiated_series = differentiated # Store first valid series

    # Return the last computed series that passed
    return differentiated